package org.protux.core;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.math.BigInteger;


/** This class plans to work around some Java limitations, specially regarding the
 * lack of  unsigned primitive types, including bytes. More optimizations will
 * come, for now, immediate support is given, just for migration purposes.
 * 
 *  The methods provided here are helper for many situations, such writing binary headers or
 *  dumping bytes from buffers. It is NOT a good idea to use theses methods on situations where performance 
 *  is critical, such mixing routinges, or FFT maths, or anything like that. FileChannels with ByteBuffers or
 *  ShortBuffers "viewrs" are the best choice in those cases.
 *  
 *  in all methods described here, "short" is being used to represent a unsigned byte (ubyte) and byte to 
 *  represent a signed byte (byte)
 *  
 *  
 *  Conventions :
 *  Legend : Description , store type, range
 *  
 *  S8  : Signed    8 bits , byte  ,   -128 .. +127
 *  U7  : Unsigned  7 bits , byte  ,      0 .. +127
 *  U8  : Unsigned  8 bits , short ,      0 .. +255
 *  S16 : Signed   16 bits , short , -32768 .. +32767
 *  U15 : Unsigned 15 bits , short ,      0 .. +32767
 *  U16 : Unsigned 16 bits , int   ,      0 .. +65535
 *  S32 : Signed   32 bits , int   ,  -2^31 .. +2^31
 */
public class Byter
	{

	//========================================================
	//               METHODS TO READ FROM BUFFERS 
	//========================================================

	/** returns a String using ubytes from a given buffer */
	public static String readString(short[] src, int start, int len)
		{
		byte b[] = new byte[len];
		for (int i = 0; i < len; i++)
			{
			b[i] = (byte) src[start + i];
			}
		return new String(b);
		}


	public static int readInt(short[] src, int start)
		{
		byte b[] = new byte[4];
		for (int i = 0; i < 4; i++)
			{
			b[i] = (byte) src[start + i];
			}
		return S8Buffer_to_Int(b);
		}


	/** returns a 0-65535 value*/
	public static final int read_U16(short[] src, int start)
		{
		int b1 = src[start];
		int b2 = src[start + 1];
		return b1 + b2 * 256;
		}


	public static final short read_S16(byte[] src, int start)
		{
		int b1 = src[start];
		int b2 = src[start + 1];
		return (short) (b1 + b2 * 256);
		}


	/** read ubytes from a file into a ubyte buffer*/
	public static void readMany_U15(RandomAccessFile rfile, short[] unsignedBuffer) throws IOException
		{
		int t = unsignedBuffer.length;
		byte signedBuffer[] = new byte[t];
		rfile.read(signedBuffer, 0, unsignedBuffer.length);
		S8Buffer_to_U8Buffer(signedBuffer, unsignedBuffer);
		}


	//========================================================
	//               METHODS TO WRITE TO BUFFERS 
	//========================================================

	/** write ubytes to a file from a ubyte buffer*/
	public static void writeMany_U15(RandomAccessFile rfile, short[] unsignedBuffer) throws IOException
		{
		int t = unsignedBuffer.length;
		byte signedBuffer[] = new byte[t];
		for (int i = 0; i < t; i++)
			{
			signedBuffer[i] = (byte) unsignedBuffer[i];
			}
		rfile.write(signedBuffer, 0, unsignedBuffer.length);
		}


	/** write an int into a ubyte buffer*/
	public static void writeInto_U15Buffer(short[] buf, int value, int startPos)
		{
		short[] valueUBytes = Byter.int_to_U8Buffer(value);
		writeInto_U15Buffer(buf, valueUBytes, startPos);
		}


	/** write an int into a ubyte buffer*/
	public static void writeInto_U15Buffer(short[] buf, long value, int startPos)
		{
		short[] valueUBytes = Byter.long_to_U8Buffer(value);
		writeInto_U15Buffer(buf, valueUBytes, startPos);
		}


	/** write an array of ubytes into another ubyte buffer*/
	public static void writeInto_U15Buffer(short[] buf, short[] ubytes, int startPos)
		{
		int len = ubytes.length;
		for (int i = 0; i < len; i++)
			{
			buf[startPos + i] = ubytes[i];
			}

		}


	/** write string into a ubyte buffer*/
	public static void writeInto_U15Buffer(short[] buf, String s, int startPos)
		{
		int len = s.length();
		for (int i = 0; i < len; i++)
			{
			char c = s.charAt(i);
			short ub = char_to_U7(c);
			buf[startPos + i] = ub;
			}
		}


	//========================================================
	//               CONVERSION METHODS 
	//========================================================

	/** convert an array of bytes into an int */
	public static int S8Buffer_to_Int(byte[] b)
		{
		int r = (b[0] & 0xFF) * 0x1000000 + (b[1] & 0xFF) * 0x10000 + (b[2] & 0xFF) * 0x100 + (b[3] & 0xFF);
		return r;
		}


	/** convert an array of ubytes into an int*/
	public static int U8Buffer_to_int(short[] b)
		{
		int r = b[0] * 0x1000000 + b[1] * 0x10000 + b[2] * 0x100 + b[3];
		return r;
		}


	/** convert/break/decouple an int into an array of bytes*/
	public static byte[] int_to_S8Buffer(int x)
		{
		byte[] b = new byte[4];
		b[0] = (byte) (x >>> 24);
		b[1] = (byte) ((x & 0x00FF0000) >>> 16);
		b[2] = (byte) ((x & 0x0000FF00) >>> 8);
		b[3] = (byte) (x & 0x000000FF);
		return b;
		}


	/** convert/break/decouple an int into an array of ubytes*/
	public static short[] int_to_U8Buffer(int x)
		{
		short[] b = new short[4];
		b[0] = (short) (x >>> 24);
		b[1] = (short) ((x & 0x00FF0000) >>> 16);
		b[2] = (short) ((x & 0x0000FF00) >>> 8);
		b[3] = (short) (x & 0x000000FF);
		return b;
		}


	/** convert/break/decouple a long into an array of ubytes*/
	public static short[] long_to_U8Buffer(long x)
		{
		byte[] b;
		BigInteger big = new BigInteger(Long.toString(x));
		b = big.toByteArray();
		short[] ubuf = new short[8];
		S8Buffer_to_U8Buffer(b, ubuf);
		return ubuf;
		}


	/** convert/break/decouple a long into an array of bytes*/
	public static byte[] int_to_S8Buffer(long x)
		{
		BigInteger big = new BigInteger(Long.toString(x));
		return big.toByteArray();
		}


	/** convert a byte into a ubyte*/
	public static short S8_to_U7(byte b)
		{
		return (short) (b & 0xFF);
		}


	/** convert a char into a ubyte*/
	public static short char_to_U7(char c)
		{
		return (short) (((byte) c) & 0xFF);
		}


	/** convert an array of ubytes to an array of bytes*/
	public static byte[] U8Buffer_to_S8Buffer(short[] src)
		{
		int t = src.length;
		byte bytes[] = new byte[t];
		for (int i = 0; i < t; i++)
			{
			bytes[i] = (byte) src[i];
			}
		return bytes;
		}


	public static void S8Buffer_to_U8Buffer(byte[] signedBuffer, short[] unsignedBuffer)
		{
		int t = signedBuffer.length;
		for (int i = 0; i < t; i++)
			{
			unsignedBuffer[i] = (short) (0x000000FF & signedBuffer[i]);
			}
		}


	/** return a 15 bits unsigned short of a given a 16bits signed short

	 * @param sho
	 * @return
	 */
	public static short S16_to_U15(short sho)
		{
		return sho;//(short) (sho & 0x000000FF); 
		}


	//========================================================
	//               METHODS TO DUMP / DEBUG  
	//========================================================

	/** return a string which the text representantion of a ubyte buffer*/
	public static String dump(short[] ubytes)
		{
		return dump(U8Buffer_to_S8Buffer(ubytes));
		}


	/** return a string which the text representantion of a byte buffer*/
	public static String dump(byte[] b)
		{
		return dump(b, b.length);
		}


	public static String dump(byte[] b, int count)
		{
		StringBuffer out = new StringBuffer("");
		int t = count;
		for (int i = 0; i < t; ++i)
			{
			if (i % 16 == 0)
				{
				out.append("" + (Integer.toHexString((i & 0xFFFF) | 0x10000).substring(1, 5) + " - "));
				}
			out.append("" + (Integer.toHexString((b[i] & 0xFF) | 0x100).substring(1, 3) + " "));
			if (i % 16 == 15 || i == b.length - 1)
				{
				int j;
				for (j = 16 - i % 16; j > 1; --j)
					out.append("" + ("   "));
				out.append("" + (" - "));
				int start = (i / 16) * 16;
				int end = (b.length < i + 1) ? b.length : (i + 1);
				for (j = start; j < end; ++j)
					if (b[j] >= 32 && b[j] <= 126)
						out.append("" + ((char) b[j]));
					else
						out.append("" + ("."));
				out.append("\n");
				}
			}
		return out.toString();
		}


	/** Unit tests*/
	/**
	 * @param args
	 */
	public static void main(String[] args)
		{
		int anInt = 239385;
		byte sByte = (byte) 129;
		System.out.println("Testing uByte(byte b):  signed=" + sByte + " unsigned=" + S8_to_U7(sByte));
		System.out.println("  ");

		System.out.println("Testing breaking int into bytes and uBytes");
		byte[] b = int_to_S8Buffer(anInt);
		short[] c = int_to_U8Buffer(anInt);
		System.out.println("int x=" + anInt);
		System.out.println("after intToBytes  :  b0=" + b[0] + "    b1=" + b[1] + "    b2=" + b[2] + "    b3=" + b[3]);
		System.out.println("after intToUBytes :  b0=" + c[0] + "    b1=" + c[1] + "    b2=" + c[2] + "    b3=" + c[3]);
		int r1 = S8Buffer_to_Int(b);
		int r2 = U8Buffer_to_int(c);
		System.out.println("after bytesToInt r=" + r1);
		System.out.println("after uBytesToInt r=" + r2);

		System.out.println("Testing longToUbytes() and longToBytes():");
		long lo = 1234567890123456789l;
		short[] ulo = long_to_U8Buffer(lo);
		byte[] slo = int_to_S8Buffer(lo);
		System.out.println("lo=" + lo);
		System.out.print("Unsigned bytes :");
		for (int i = 0; i < 8; i++)
			System.out.print("b" + i + "=" + ulo[i] + "  ");
		System.out.println("");
		System.out.print("Signed bytes   :");
		for (int i = 0; i < 8; i++)
			System.out.print("b" + i + "=" + slo[i] + "  ");
		System.out.println("");

		System.out.println("Testing readUShort(byte[],pos)");
		byte[] t4 = { -10, -4, 20, 1 };
		short shox = read_S16(t4, 0);
		System.out.println("from byte[] read " + shox + " at position 0");
		System.out.println(" ");
		System.out.println(" ");

		}

	}
