package org.protux.core;

import static org.protux.core.Debugger.PERROR;
import static org.protux.core.Debugger.PMESG;
import static org.protux.core.Debugger.PWARN;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.protux.core.editor.Song;
import org.protux.core.media.audio.AudioMedia;
import org.protux.core.media.audio.Ogg;
import org.protux.core.media.audio.Praf;
import org.protux.core.media.audio.Wav;


public class Factory
	{
	protected static String mp3DecodeCommand = "mpg123 -w %of %if";

	protected static String oggDecodeCommand = "ogg123 -d wav -f %of %if";

	protected static String oggEncodeCommand = "ogg123 -d wav -f %of %if";

	public static final int MEDIA_FORMAT_PRAF = 0;

	public static final int MEDIA_FORMAT_OGG = 1;

	public static final int MEDIA_FORMAT_WAV = 2;

	public static final int MEDIA_FORMAT_MP3 = 3;

	public static final int MEDIA_FORMAT_AVI = 4;


	public static void init()
		{
		set_mp3_decode_command(GlobalProperties.get(GlobalProperties.KEY_MP3_DECODE_COMMAND));
		set_ogg_decode_command(GlobalProperties.get(GlobalProperties.KEY_OGG_DECODE_COMMAND));
		set_ogg_encode_command(GlobalProperties.get(GlobalProperties.KEY_OGG_ENCODE_COMMAND));
		}


	/** try to probe media format.. not the best way.. need improvements */
	public static int probeMediaFormat(File file)
		{
		String pureName = file.getName();
		int x = pureName.lastIndexOf(".") + 1;
		String ext = pureName.substring(x);
		if (ext.equalsIgnoreCase("wav"))
			return MEDIA_FORMAT_WAV;
		else if (ext.equalsIgnoreCase("praf"))
			return MEDIA_FORMAT_PRAF;
		return -1;
		}


	public static void set_mp3_decode_command(String com)
		{
		mp3DecodeCommand = com;
		}


	public static void set_ogg_decode_command(String com)
		{
		oggDecodeCommand = com;
		}


	public static void set_ogg_encode_command(String com)
		{
		oggEncodeCommand = com;
		}


	/**
	 * given a project and a song ID (number), try to load the song, or create
	 * in case the song for that id does not exist
	 */
	public static Song createOrOpenSong(Project project, int songNumber) throws IOException
		{
		Song song = new Song(project, songNumber);
		boolean newSong = false;
		try
			{
			song.load();
			}
		catch (IOException ex)
			{
			newSong = true;
			}
		if (newSong)
			{
			int numTracks = GlobalProperties.getInt(GlobalProperties.KEY_DEFAULT_NUM_TRACKS);
			int rate = GlobalProperties.getInt(GlobalProperties.KEY_DEFAULT_SAMPLE_RATE);
			int bitDepth = GlobalProperties.getInt(GlobalProperties.KEY_DEFAULT_BIT_DEPTH);
			song.rate = rate;
			song.bitDepth = bitDepth;
			song.artists = "No Artists Defined";
			song.editingMode = Const.EDITING_MODE_NORMAL;
			song.hzoom = GlobalProperties.getInt(GlobalProperties.KEY_HZOOM_LEVEL);
			for (int i = 0; i < numTracks; i++)
				{
				song.addTrack();
				}
			File cRootDir = new File(song.rootDir);
			if (!cRootDir.exists())
				{
				if (!cRootDir.mkdirs())
					{
					PERROR("Cannot create dir " + song.rootDir);
					}
				}
			File cAudioSourcesDir = new File(song.audioSourcesDir);
			if (!cAudioSourcesDir.exists())
				{
				if (!cAudioSourcesDir.mkdirs())
					{
					PERROR("Cannot create dir  " + song.audioSourcesDir);
					}
				}
			File cCaptureDir = new File(song.captureDir);
			if (!cCaptureDir.exists())
				{
				if (!cCaptureDir.mkdirs())
					{
					PERROR("Cannot create dir " + song.captureDir);
					}
				}
			song.save();
			}
		return song;
		}


	/** self explanable huh ... */
	public static Project createProject(String projectName, int numSongs) throws IOException
		{
		log("Creating new project " + projectName + "  NumSongs=" + numSongs);
		if (numSongs < 1)
			{
			PWARN("Invalid number of songs requested for this new project (" + numSongs + ") . Assuming 1 (one).");
			numSongs = 1;
			}
		Project project = new Project(projectName);
		File f = new File(project.getRootDir());
		if (!f.exists() && !f.mkdirs())
			{
			throw new IOException("Project.create() : Cannot create project path " + project.getRootDir());
			}
		project.resetProperties();
		for (int i = 0; i < numSongs; i++)
			{
			int songNumber = project.getNumSongs();
			project.song[i] = Factory.createOrOpenSong(project, songNumber);
			}
		project.setNumSongs(numSongs);
		project.setCurrentSong(0);
		String msg = "Project " + project.getTitle() + " created.";
		Protux.info(msg);
		log(msg);
		return project;
		}


	/**
	 * render a copy of the given file into a new PRAF , under the destination
	 * dir
	 */
	public static Praf importTo(File srcFile, Song song) throws IOException
		{
		String destinationDir = song.audioSourcesDir;
		String srcFilename = srcFile.getName();
		String targetFilename = destinationDir + "/" + srcFilename.substring(0, srcFilename.indexOf(".")) + ".praf";
		Praf praf = new Praf(new File(targetFilename),Const.PEAK_LESS);
		int format = Factory.probeMediaFormat(srcFile);

		if (format == Factory.MEDIA_FORMAT_PRAF)
			{
			log("Copying PRAF : " + srcFile.getPath() + " to " + targetFilename);
			praf = (Praf) (new Praf(srcFile)).copyTo(new File(targetFilename));
			}
		if (format == Factory.MEDIA_FORMAT_WAV)
			{
			log("Importing WAV : " + srcFile.getPath() + " to " + targetFilename);
			Wav wav = new Wav(srcFile, Const.PEAK_LESS, Const.READ_ONLY);
			wav.readHeader();
			String t = song.parentProject.getTitle();
			SimpleDateFormat d = new SimpleDateFormat("yyyy-MM-dd");
			String date = "in " + d.format(new Date(System.currentTimeMillis()));
			praf.set(wav.rate, wav.channels, wav.bitDepth, t, "Imported Media", "Generic", date);
			praf.writeHeader();
			praf.appendSamplesFromFile(wav);
			praf.fixAudioSize();
			}
		else if (format == Factory.MEDIA_FORMAT_MP3)
			{
			/*
			 * MIG String decodeCommand; if (filename.find("mp3", 0, false) >=
			 * 0) { decodeCommand = mp3DecodeCommand; } else { decodeCommand =
			 * oggDecodeCommand; } if ((decodeCommand.find("%if") < 0) ||
			 * (decodeCommand.find("%of") < 0)) { String msg = "Decode Command
			 * is invalid. Please change it. It must use %if and %of to place
			 * input file and output file references , respectively. Exemple :
			 * mpg123 -w %of %if"; //PERROR(msg); } String inFilename =
			 * originalName; String outFilename; if (filename.find("mp3"
			 * ,0,false)>=0) { outFilename= destinationDir + "/" +
			 * pureName.replace(QRegExp(".mp3"),".wav"); } else { outFilename=
			 * destinationDir + "/" + pureName.replace(QRegExp(".ogg"),".wav"); }
			 * decodeCommand =
			 * decodeCommand.replace(QRegExp("%if"),"\""+inFilename+"\"");
			 * decodeCommand =
			 * decodeCommand.replace(QRegExp("%of"),"\""+outFilename+"\"");
			 * System.out.println("\n"); //PWARN("Executing External command :
			 * %s\n", decodeCommand); System.out.println("Begin of output :\n");
			 * System.out.println("----------------------------------------\n");
			 * File pipe_in; final String ccommand = decodeCommand; //MIG
			 * pipe_in = popen (ccommand, "r"); if (false)//MIG !pipe_in) { //
			 * TODO Jan error message ... //PEXIT; return -1; } char c; //MIG
			 * while (!feof(pipe_in)) //MIG { //MIG fread(&c,1,1,pipe_in); //MIG }
			 * //MIG pclose (pipe_in);
			 * System.out.println("----------------------------------------\n");
			 * System.out.println("End of output.\n\n"); if (false)//MIG
			 * filename.find("mp3" ,0,false)>=0) { //MIG pureName =
			 * pureName.replace(QRegExp(".mp3"),".wav"); } else { //MIG pureName =
			 * pureName.replace(QRegExp(".ogg"),".wav"); }
			 */
			}
		else
			{
			throw new IOException("Unkown Audio Format");
			}
		praf.setPeakless(false);
		return praf;
		}


	/** Just a helper to quick OPEN a media (automatically determining its type) */
	public static AudioMedia getMedia(File file) throws IOException
		{
		int format = Factory.probeMediaFormat(file);
		switch (format)
			{
			case Factory.MEDIA_FORMAT_OGG:
				{
				Ogg oggFile = new Ogg(file);
				oggFile.readHeader();
				oggFile.skipHeader();
				oggFile.closeFile();
				return oggFile;
				}
			case Factory.MEDIA_FORMAT_PRAF:
				{
				Praf prafFile = new Praf(file);
				prafFile.readHeader();
				prafFile.skipHeader();
				prafFile.closeFile();
				return prafFile;
				}
			case Factory.MEDIA_FORMAT_WAV:
				{
				Wav wavFile = new Wav(file);
				wavFile.readHeader();
				wavFile.skipHeader();
				wavFile.closeFile();
				}
			default:
			PERROR("no known filetype :-(");
			}

		return null;
		}


	/**
	 * Just a helper to create CREATE a new media, with rate, channels and
	 * bitdepth specified
	 */
	public static AudioMedia createMedia(File file, int format, int channels, int rate, int bitDepth, String project, String group, String hardware, String description) throws IOException
		{
		AudioMedia outFile = null;
		switch (format)
			{
			case Factory.MEDIA_FORMAT_PRAF:
			outFile = new Praf(file);
			break;
			case Factory.MEDIA_FORMAT_WAV:
			outFile = new Wav(file);
			break;
			default:
				{
				PERROR("Format not supported");
				return null;
				}
			}
		outFile.set(channels, rate, bitDepth, project, group, hardware, description);
		outFile.writeHeader();
		outFile.skipHeader();
		return outFile;
		}


	private static void log(String m)
		{
		PMESG(2, "[ Factory ] " + m);
		}

	}
