package org.protux.core.editor;

import static org.protux.core.Debugger.PERROR;
import static org.protux.core.Debugger.PMESG;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.ShortBuffer;
import java.nio.channels.FileChannel;

import org.protux.core.Const;
import org.protux.core.exceptions.PeakException;
import org.protux.core.media.audio.AudioMedia;


/** 
 * A Peak is a class that aggregates collections of n Audio Peaks Representations for Zoomming
 * Each Zoom Image is a buffer for a given zoom level 
 * each buffer contains n blocks ( actual plots in the screen ) for that zoom level. 
 * The number of blocks will vary acording to the media size. a media of 1000 samples will have 100 visible blocksin 10:1 zoom level
 * each Block has n peaks, being n the number of channels for the associated media. A stereo file will have 2 peaks for each block (left and right channel)
 * each peak represents the actual sample value , using a S16 normalized value (See Byter.java), independently of sample size, bit depth, whatever
 * in the original media.
 * 
 * Up to a certain zoom level, peaks are read on the fly from the hard disk (ONLINE PEAKS), and above that level , 
 * they are pre-cached into RAMBuffers, one for each zoom level (CACHED PEAKS)
 * 
 */

public final class Peak
	{
	public static final int NOT_AVAILABLE = 0;
	public static final int BUILDING = 1;
	public static final int AVAILABLE = 2;
	public static final int BLOCKED = 4;

	protected AudioMedia sourceMedia;

	protected final ShortBuffer[] zoomImageBuffer = new ShortBuffer[Const.MAX_ZOOM_LEVELS]; //S16 values, one for each sample
	protected final int[] zoomImageBufferStatus = new int[Const.MAX_ZOOM_LEVELS];
	protected final int[] zoomImageBufferBuildProgress = new int[Const.MAX_ZOOM_LEVELS];

	protected final PeakBuildThread[] peakBuildThread = new PeakBuildThread[Const.MAX_ZOOM_LEVELS];

	protected int totalZoomImagesSize;
	protected boolean sourceMediaDefined = false;
	protected final int bufferingAccuracy = 4; // 1, 2, 4, 8, 16, 32, 64, 128 (the more the number, the less accuracy/fidelity in the peak representation)
	protected final float scanFidelity = 1.0f; // 0.50-1.00 value, indicate how much of chunck of data to be analised should be read... 100% reads the full range to elect a sample, 50% read the half left portion... 
	private boolean threaded = false;
	private final short electedStereoBlock[] = new short[2];


	public Peak ( )
		{
		sourceMediaDefined = false;
		}


	public Peak (AudioMedia anAudioSource) throws PeakException
		{
		defineSourceMedia(anAudioSource);
		}


	public void defineSourceMedia(AudioMedia source) throws PeakException
		{
		if (sourceMediaDefined)
			{
			return;
			}
		this.sourceMedia = source;
		this.sourceMediaDefined = true;
		createRAMBuffers();
		}


	public void createRAMBuffers() throws PeakException
		{
		if (sourceMediaDefined)
			{
			long totalBlocks = sourceMedia.totalBlocks;
			int blockSize = sourceMedia.blockSize;
			int channels = sourceMedia.channels;
			int tot = 0;
			for (int hzoom = Const.MAX_ZOOM_USING_SOURCEFILE + 1; hzoom < Const.MAX_ZOOM_LEVELS; hzoom++, tot++)
				{
				int zs = Const.ZOOM_STEP[hzoom];
				int numBlocksForThisZoomLevel = (int) totalBlocks / zs; // it MUST fit into an int... we suppose that a peak image will never get more than 2^31 steps
				int zoomImageSize = numBlocksForThisZoomLevel * blockSize;
				ByteBuffer bb = ByteBuffer.allocate(zoomImageSize);
				bb.position(0);
				bb.order(ByteOrder.LITTLE_ENDIAN);
				zoomImageBuffer[hzoom] = bb.asShortBuffer();
				zoomImageBufferStatus[hzoom] = NOT_AVAILABLE;
				totalZoomImagesSize += zoomImageSize;
				}
			log("Created " + tot + " RAM buffers. Total RAM Size allocated : " + totalZoomImagesSize);
			}
		}


	/**
	 * 
	 * @param bufferToFill
	 * @param startBlock
	 * @param numBlocks
	 * @param hzoom
	 * @return the number of blocks filled
	 */
	public int fillBuffer(short[] bufferToFill, long startBlock, int numBlocks, int hzoom)
		{
		int blocksFilled = 0;
		int zs = Const.ZOOM_STEP[hzoom];
		if (hzoom > Const.MAX_ZOOM_USING_SOURCEFILE)
			{
			int startPosInZoomImage = (int) (startBlock * sourceMedia.blockSize) / zs;
			int sampleCount = numBlocks * sourceMedia.channels;
			zoomImageBuffer[hzoom].position(startPosInZoomImage);
			zoomImageBuffer[hzoom].get(bufferToFill, 0, sampleCount);
			blocksFilled = numBlocks;
			}
		else
			{ // READ PEAKS DIRECT FROM DISK 
			blocksFilled = 0;
			try
				{
				long originalPos = sourceMedia.byteTell();
				long startBytePos = sourceMedia.headerSize + startBlock * sourceMedia.blockSize;
				final int stepInShorts = zs * sourceMedia.channels;
				int totalShorts = numBlocks * sourceMedia.channels; 
				short electionBuf[] = new short[stepInShorts];
				int p=0;
				FileChannel fc = sourceMedia.getFile().getChannel();
				fc.position(startBytePos);
				ByteBuffer diskPeakBuffer = ByteBuffer.allocate(numBlocks * sourceMedia.blockSize * zs).order(ByteOrder.LITTLE_ENDIAN);
				diskPeakBuffer.position(0);
				ShortBuffer diskPeakBufferShortView = diskPeakBuffer.asShortBuffer();
				diskPeakBufferShortView.position(0);
				fc.read(diskPeakBuffer);
				if (sourceMedia.channels == Const.STEREO)
					{
					for (int i = 0; i < totalShorts; i += 2 )
						{
						diskPeakBufferShortView.get(electionBuf);
						short[] electeds = electStereo(electionBuf);
						bufferToFill[i] = electeds[0];
						bufferToFill[i + 1] = electeds[1];
						p += stepInShorts;
						diskPeakBufferShortView.position(p);
						}
					}
				else
					{
					for (int i = 0 ; i < totalShorts; i++)
						{
						diskPeakBufferShortView.position(p);
						diskPeakBufferShortView.get(electionBuf);
						bufferToFill[i] = electMono(diskPeakBufferShortView);
						p += stepInShorts;
						}
					}
				sourceMedia.byteSeek(originalPos);
				blocksFilled = numBlocks;
				}
			catch (IOException e)
				{
				e.printStackTrace();
				}
			}
		return blocksFilled;
		}



	private final short[] electStereo(ShortBuffer electionBuf)
		{
		int t = electionBuf.capacity();
		short electedL = 0,electedR = 0,shoL = 0,shoR = 0;
		electionBuf.position(0);
		for (int i = 0; i < t; i+=2) // FIXME USar accuracy e amostrage factors
			{
			shoL = electionBuf.get();
			shoR = electionBuf.get();
			electedL = (Math.abs(shoL) > Math.abs(electedL)) ? shoL : electedL;
			electedR = (Math.abs(shoR) > Math.abs(electedR)) ? shoR : electedR;
			}
		electedStereoBlock[0] = electedL;
		electedStereoBlock[1] = electedR;
		return electedStereoBlock;
		}

	private final short[] electStereo(short[] electionBuf)
		{
		int t = electionBuf.length;
		short electedL = 0,electedR = 0,shoL = 0,shoR = 0;
		for (int i = 0; i < t; i+=2) // FIXME USar accuracy e amostrage factors
			{
			shoL = electionBuf[i];
			shoR = electionBuf[i+1];
			electedL = (Math.abs(shoL) > Math.abs(electedL)) ? shoL : electedL;
			electedR = (Math.abs(shoR) > Math.abs(electedR)) ? shoR : electedR;
			}
		electedStereoBlock[0] = electedL;
		electedStereoBlock[1] = electedR;
		return electedStereoBlock;
		}

	private final short electMono(ShortBuffer electionBuf)
		{
		int t = electionBuf.capacity();
		short elected = 0,sho = 0;
		electionBuf.position(0);
		for (int i = 0; i < t; i++) // FIXME USar accuracy e amostrage factors
			{
			sho = electionBuf.get();
			elected = (Math.abs(sho) > Math.abs(elected)) ? sho : elected;
			}
		return elected;
		}


	public int getBuildStatus(int hzoom)
		{
		int buildStatus = zoomImageBufferStatus[hzoom];
		if (buildStatus == NOT_AVAILABLE)
			{
			buildStatus = startBuild(hzoom); // start the build immediately
			}
		return buildStatus;
		}


	private int startBuild(int hzoom)
		{
		int r = BUILDING;
		if (threaded)
			{
			peakBuildThread[hzoom] = new PeakBuildThread(hzoom);
			peakBuildThread[hzoom].start();
			}
		else
			{
			r = buildImageForZoom(hzoom);
			}
		return r;
		}


	public int getBuildingProgress(int hzoom)
		{
		return zoomImageBufferBuildProgress[hzoom];
		}


	protected int buildImageForZoom(int hzoom)
		{
		if (hzoom <= Const.MAX_ZOOM_USING_SOURCEFILE)
			{
			zoomImageBufferStatus[hzoom] = AVAILABLE;
			return AVAILABLE;
			}
		zoomImageBufferStatus[hzoom] = BUILDING;
		try
			{
			sourceMedia.rewindAudio();
			if (!loadPreviousSavedImage())
				{
				if (sourceMedia.channels == Const.MONO)
					scanMono(hzoom);
				else
					scanStereo(hzoom);
				saveImage();
				}
			zoomImageBufferStatus[hzoom] = AVAILABLE;
			}
		catch (Exception e)
			{
			log("Could not build peak image for hzoom=" + hzoom + " : " + e);
			e.printStackTrace();
			zoomImageBufferStatus[hzoom] = BLOCKED;
			return BLOCKED;
			}
		return AVAILABLE;
		}


	private void scanMono(int hzoom) throws IOException
		{
		RandomAccessFile rf = sourceMedia.getFile();
		final int zs = Const.ZOOM_STEP[hzoom];
		final int blockSize = sourceMedia.blockSize;
		final long totalBlocksInSource = sourceMedia.totalBlocks;
		long blockPosInSource = sourceMedia.getFirstAudioBytePos();
		ShortBuffer zoomImage = zoomImageBuffer[hzoom];
		zoomImage.position(0);
		FileChannel fc = rf.getChannel();
		int amostrageSize = zs * blockSize;
		ByteBuffer amoReadBuf = ByteBuffer.allocate(amostrageSize);
		amoReadBuf.order(ByteOrder.LITTLE_ENDIAN);
		ShortBuffer amoElecBuf = amoReadBuf.asShortBuffer();
		while (blockPosInSource < totalBlocksInSource)
			{
			fc.position(blockPosInSource * blockSize);
			amoReadBuf.position(0);
			amoElecBuf.position(0);
			fc.read(amoReadBuf);
			zoomImage.put(electMono(amoElecBuf));
			blockPosInSource += zs;
			int prog = (int) (blockPosInSource / totalBlocksInSource * 100);
			zoomImageBufferBuildProgress[hzoom] = prog;
			}
		amoReadBuf = null;
		amoElecBuf = null;
		}


	private void scanStereo(int hzoom) throws IOException
		{
		RandomAccessFile rf = sourceMedia.getFile();
		int zs = Const.ZOOM_STEP[hzoom];
		final int blockSize = sourceMedia.blockSize;
		ShortBuffer zoomImage = zoomImageBuffer[hzoom];
		zoomImage.position(0);
		sourceMedia.rewindAudio();
		FileChannel fc = rf.getChannel();
		int amostrageSize = zs * blockSize;
		ByteBuffer amoReadBuf = ByteBuffer.allocate(amostrageSize);
		amoReadBuf.position(0);
		amoReadBuf.order(ByteOrder.LITTLE_ENDIAN);
		ShortBuffer amoReadBufferShortView = amoReadBuf.asShortBuffer();
		int numBlocksForThisZoomLevel = zoomImage.capacity();
		long startBytePos = sourceMedia.getFirstAudioBytePos();
		for (int i = 0; i < numBlocksForThisZoomLevel; i+=2)
			{
			amoReadBuf.position(0);
			fc.read(amoReadBuf,startBytePos);
			short[] electeds = electStereo(amoReadBufferShortView);
			zoomImage.put(electeds[0]);
			zoomImage.put(electeds[1]);
			startBytePos+=amostrageSize;
			int prog = (int) ( i / (numBlocksForThisZoomLevel * 100.0f));
			zoomImageBufferBuildProgress[hzoom] = prog;
			}
		}


	public void sourceMediaChanged()
		{
		for (int hzoom = Const.MAX_ZOOM_USING_SOURCEFILE + 1; hzoom < Const.MAX_ZOOM_LEVELS; hzoom++)
			{
			zoomImageBuffer[hzoom] = null;
			zoomImageBufferStatus[hzoom] = BLOCKED;
			totalZoomImagesSize = 0;
			}
		}


	public static void log(String m)
		{
		PMESG(3, "[ Peak ] " + m);
		}


	public void process_samples_from_buffer(byte[] buffer, int bufferSize)
		{
		// MIG I THINK THIS WHOLE METHOD CAN BE A SIMPLE Arrays.fill/copy CALL
		/* MIG 
		 int i = 0;
		 short sho = 0;
		 int posBuffer = 0;
		 while (i < bufferSize)
		 {
		 for (int j = 0; j < sourceMedia.channels; j++)
		 {
		 //MIG byte[] p = posBuffer + j*SIZE_OF_SHORT;
		 //MIG sho = *(short*)p;
		 sho /= 256;
		 // MIG i += SIZE_OF_SHORT;
		 if (sho < 0)
		 sho *= -1;
		 }
		 //			 MIG posBuffer += SIZE_OF_SHORT * sourceMedia.channels;
		 }*/
		}


	public void saveImage() throws IOException
		{
		/*
		 // MIG I THINK THIS WHOLE METHOD CAN BE A SIMPLE Arrays.fill/copy CALL  with file.writeBytes
		 String peakFilename = sourceMedia.getRootDir() + "/" + sourceMedia.get_pure_name() + ".peak";
		 log("Saving Peak to " + peakFilename);
		 File file = new File(peakFilename);
		 byte[] buf;
		 //store the values from RAMBuffer in a (temp) buf, and write buf at ones to the hard disk, //delete buf afterwards.
		 buf = new byte[totalRAMBufferSize * SIZE_OF_CHAR]; // MIG  IS THIS CORRECT ? totalRAMBUfferSize SHOULDNT BE ALREADY THE SIZE OF BUFFER , SO, NO NEED THIS * SIZE_OF_CHAR
		 byte[] p = buf;
		 for (int hzoom = MAX_ZOOM_USING_SOURCEFILE + 1; hzoom < MAX_ZOOM_LEVELS; hzoom++)
		 {
		 for (int j = 0; j < RAMBufferSize[hzoom]; j++)
		 {
		 byte sho = RAMBuffer[hzoom][j];
		 // MIG *(p++) = sho;
		 }
		 }
		 // MIG fwrite(buf,SIZE_OF_CHAR,totalRAMBufferSize,file);
		 log("RAM Buffers saved. Total size of file is " + (totalRAMBufferSize * SIZE_OF_CHAR));
		 // MIG file.close();
		 buf=null;
		 */
		}


	public boolean loadPreviousSavedImage() throws IOException
		{
		/*
		 // MIG I THINK THIS WHOLE METHOD CAN BE A SIMPLE Arrays.fill/copy CALL with file.readBytes
		 String peakFilename = sourceMedia.getRootDir() + "/" + sourceMedia.get_pure_name() + ".peak";
		 log("Looking for peak image file " + peakFilename);
		 File pfile = new File(peakFilename);
		 byte[] buf;
		 //reading peakfile into (temp) buf, so the file reading can be done in one big slurp.
		 buf = new byte[totalRAMBufferSize * SIZE_OF_CHAR];
		 //MIG fread(buf,SIZE_OF_CHAR,totalRAMBufferSize,pfile);
		 byte[] p = buf;
		 //putting the values in the right manner into RAMBuffer, deleting buf afterwards
		 for (int hzoom = MAX_ZOOM_USING_SOURCEFILE + 1; hzoom < MAX_ZOOM_LEVELS; hzoom++)
		 {
		 //MIG memcpy(RAMBuffer[hzoom], p, RAMBufferSize[hzoom] * SIZE_OF_CHAR);
		 //MIG p += RAMBufferSize[hzoom] * SIZE_OF_CHAR;
		 }

		 log("RAM Buffers sucessfully loaded.");
		 //MIG fclose(pfile);
		 //delete buf;
		 */
		return false;
		}


	//	 ======================= Threaded On-Demand Peak Generation ==================================
	//	 Each Peak Zoom Image will be generated whenever its displaying is required for the first time.
	//	 That will happen also when changing zoom. If the image for a given zoom is not available yet,
	//	 (RAMBufferBuilt[zoom] is false), it will show "Building peaks" and then the peak becomes
	//	 available.
	//	 =============================================================================================
	private class PeakBuildThread extends Thread
		{
		private int hzoom = -1;


		PeakBuildThread (int hzoom)
			{
			this.hzoom = hzoom;
			}


		@Override
		public void run()
			{
			try
				{
				buildImageForZoom(hzoom);
				}
			catch (Exception ex)
				{
				PERROR("Error : " + ex);
				}
			}

		}

	}
/* AVERAGE TECHINIQUE 
 int t = electionBuf.length;
 int tblocks = t/2;
 int electedL = 0;
 int electedR = 0;
 for (int i = 0; i < t; i += 2)
 {
 electedL += electionBuf[i];
 electedR += electionBuf[i + 1];
 }
 electedSamples[0] = (short) (electedL/tblocks);
 electedSamples[1] = (short) (electedR/tblocks);
 */
