package org.protux.core.editor;

import static org.protux.core.Debugger.PERROR;
import static org.protux.core.Debugger.PMESG;
import static org.protux.core.gui.ColorManager.CM_TRACK_BG;
import static org.protux.core.gui.ColorManager.CM_TRACK_BG_ACTIVE;
import static org.protux.core.gui.ColorManager.CM_TRACK_PANEL_BG;
import static org.protux.core.gui.ColorManager.getCmColor;
import static org.protux.core.gui.ColorManager.setCmBlur;

import java.awt.Color;
import java.awt.Graphics2D;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;

import org.protux.core.Const;
import org.protux.core.Protux;
import org.protux.core.SchemaParser;
import org.protux.core.engines.audio.SoftBus;
import org.protux.core.engines.audio.SoftBusMapper;
import org.protux.core.exceptions.PeakException;
import org.protux.core.media.audio.AudioMedia;

public class Track
	{

		// TODO : ALL lists stuff should be a single super class.. for example clips, regions, selections, and so on...
		public static final int SEPARATOR_HEIGHT = 2;

		private static final Color CM_TRACK_SEPARATOR_MODE_CURVES_COLOR1 = new Color(50, 50, 50);
		private static final Color CM_TRACK_SEPARATOR_MODE_CURVES_COLOR2 = new Color(150, 150, 160);
		private static final Color CM_TRACK_SEPARATOR_MODE_FILTERS_COLOR1 = new Color(60, 60, 60);
		private static final Color CM_TRACK_SEPARATOR_MODE_FILTERS_COLOR2 = new Color(240, 240, 240);

		public LinkedList<AudioClip> clipList = new LinkedList<AudioClip>();
		public Song parentSong;
		public Chain controllerChain = new Chain(this);
		public AudioMedia recordingAudio = null;
		public AudioClip lastRecordedClip = null;
		public AudioMedia recordingAudioFile = null;
		public Editor proxyEditor;
		public Rack proxyRack;
		public SoftBusMapper proxyMixer;
		public TrackPanel panel;

		public String name;
		public float inputLevelMax[] = new float[2]; // review me
		public float inputLevel[] = new float[2]; // review me
		public float gain = Const.DEFAULT_GAIN;
		public float pan = 0f;
		public SoftBus busIn = null;
		public SoftBus busOut = null;
		public int whichChannels = Const.STEREO; // could be ONLY_LEFT_CHANNEL or ONLY_RIGHT_CHANNEL;
		public String recordingAudioFilename = null;
		public long recordingAudioStartPos = 0;
		public long recordingAudioPos = 0;
		public int recordingBufferSize = 0;
		public byte[] recordingBuffer;

		int trackNumber = 0;
		int baseY = 0;
		int height = Const.STANDARD_TRACK_HEIGHT;
		boolean isActive = false;

		public Track(Song parentSong, int trackNumber, String name)
			{
			this.parentSong = parentSong;
			this.proxyEditor = parentSong.proxyEditor;
			this.proxyRack = parentSong.proxyRack;
			this.proxyMixer = parentSong.softMapper;
			this.name = name;
			this.trackNumber = trackNumber;
			this.height = Const.STANDARD_TRACK_HEIGHT;
			this.baseY = this.trackNumber * this.height;
			this.busIn = Protux.getDefaultSoftBusIn();
			this.busOut = Protux.getDefaultSoftBusOut();
			this.panel = new TrackPanel(this);
			}


		public void recreateClipArea()
			{
			Graphics2D painter = proxyEditor.getGBuffer();
			recreateClipArea(painter);
			}


		public void recreateClipArea(Graphics2D painter)
			{
			for (AudioClip c : clipList)
				c.peaksDrawn = false;
			clearClipArea(painter);
			drawSeparator(painter);
			draw_regions(painter);
			drawClips(painter);
			repaintClipArea();
			}


		public void repaintClipArea()
			{
			proxyEditor.repaint(0, getRealBaseY(), proxyEditor.getWidth(), height);
			}


		public void clearClipArea(Graphics2D painter)
			{
			painter.setColor(isActive ? getCmColor(CM_TRACK_BG_ACTIVE) : getCmColor(CM_TRACK_BG));
			painter.fillRect(0, getRealBaseY(), proxyEditor.getWidth(), height);
			}


		public void drawSeparator(Graphics2D painter)
			{
			int by = getRealBaseY();
			int w = proxyEditor.getWidth();
			if (parentSong.editingMode == Const.EDITING_MODE_CURVES)
				{
				painter.setColor(CM_TRACK_SEPARATOR_MODE_CURVES_COLOR1);
				painter.drawLine(0, by + height, w, getRealBaseY() + height);
				painter.setColor(CM_TRACK_SEPARATOR_MODE_CURVES_COLOR2);
				painter.drawLine(0, by + height - 1, proxyEditor.getWidth(), by + height - 1);
				}
			else
				{
				painter.setColor(CM_TRACK_SEPARATOR_MODE_FILTERS_COLOR1);
				painter.drawLine(0, by + height, proxyEditor.getWidth(), by + height);
				painter.setColor(CM_TRACK_SEPARATOR_MODE_FILTERS_COLOR2);
				painter.drawLine(0, by + height - 1, proxyEditor.getWidth(), by + height - 1);
				}
			}


		public void drawClips(Graphics2D painter)
			{
			for (AudioClip c : clipList)
				{
				if (c.isVisible())
					c.draw();
				}
			if (parentSong.editingMode == Const.EDITING_MODE_NORMAL)
				{
				controllerChain.draw();
				}
			}


		public void draw_regions(Graphics2D g)
			{
			int w = proxyEditor.getClipAreaWidth();
			g.setColor(new Color(155, 100, 250));
			for (Region m : parentSong.regionList)
				{
				int xrs = parentSong.blockToEditorX(m.beginBlock);
				int xre = parentSong.blockToEditorX(m.endBlock);
				int xs, xe = 0;
				if ((xrs >= 0) && (xrs < w))
					xs = xrs;
				else if (xrs >= 0)
					xs = w;
				else
					xs = 0;
				if ((xre >= 0) && (xre < w))
					xe = xre;
				else if (xre >= 0)
					xe = w;
				else
					xe = 0;
				g.fillRect(xs, getRealBaseY(), xe - xs, 3);
				proxyEditor.repaint(xs, getRealBaseY(), xe - xs, 3);
				}
			}


		public boolean is_pointed(int y)
			{
			boolean b = ((y >= getRealBaseY()) && (y <= getRealBaseY() + height));
			return b;
			}


		/** split the clip at the given point, creating a new one STILL NOT INSERTED ON THE TRACK. The responsability to inserted the new clip is clients */
		public AudioClip splitClip(long splitPoint)
			{
			AudioClip newRightClip = null;
			AudioClip clipToBeSplit = getClipUnderBlock(splitPoint);
			if (clipToBeSplit != null)
				newRightClip = clipToBeSplit.split(splitPoint);
			return newRightClip;
			}


		public AudioClip getClipUnderBlock(long blockPos)
			{
			PMESG(5, "Trying to find clip under " + blockPos);
			AudioClip theClip = null;
			for (AudioClip c : clipList)
				{
				long fb = c.trackFirstBlock;
				long lb = c.trackLastBlock;
				if ((fb <= blockPos) && (lb >= blockPos))
					{
					theClip = c;
					break;
					}
				}
			return theClip;
			}


		public AudioClip get_clip_between(long blockPosL, long blockPosR)
			{
			AudioClip theClip = null;
			for (AudioClip c : clipList)
				{
				long fb = c.trackFirstBlock;
				long lb = c.trackLastBlock;
				if ((fb >= blockPosL) && (lb <= blockPosR))
					{
					theClip = c;
					break;
					}
				if (fb > blockPosR)
					break;
				}
			return theClip;
			}


		public AudioClip getClipAfter(long blockPos)
			{
			for (AudioClip c : clipList)
				{
				if (c.trackFirstBlock > blockPos)
					return c;
				}
			return null;
			}


		public AudioClip getClipBefore(long blockPos)
			{
			for (AudioClip c : clipList)
				{
				if (c.trackFirstBlock < blockPos)
					return c;
				}
			return null;
			}


		// FIXME needs to be reviewed
		public int delete_clip(AudioClip clip, boolean permanently)
			{
			if (clip == null)
				{
				PERROR("Trying to //delete invalid Clip");
				return -1;
				}
			if (permanently && (parentSong.getClipsCountForAudio(clip.audioSource) == 1)) // this clip is the only one with this audioSource,
				{
				// clip.audioSource.peak.unbuild(); //so we have to unbuild the peak RAMBuffers... or not? Waste of memory usage if we keep it there...
				}
			return 1;
			}


		public int get_id()
			{
			return trackNumber;
			}


		public void removeClip(AudioClip clip)
			{
			AudioClip next = clip.next;
			AudioClip prev = clip.prev;
			if (prev != null)
				prev.next = next;
			if (next != null)
				next.prev = prev;
			clipList.remove(clip);
			}


		public void addClip(AudioClip clip)
			{
			clip.setParent(this);
			long clipLenght = clip.sourceLastBlock - clip.sourceFirstBlock;
			log("Adding clip :  [ TFB = " + clip.trackFirstBlock + " , TLB = " + clip.trackLastBlock + " , SFB = " + clip.sourceFirstBlock + " , SLB = " + clip.sourceLastBlock + " , LEN = " + clipLenght + " ]");
			if (clipList.size() == 0)
				{
				clipList.add(clip);
				clip.prev = null;
				clip.next = null;
				}
			else
				{
				int i = 0;
				AudioClip lastClipBeforeNew = null;
				for (AudioClip c : clipList)
					{
					if (c.trackFirstBlock >= clip.trackFirstBlock)
						break;
					lastClipBeforeNew = c;
					i++;
					}
				log("Adding to the list");
				clip.prev = lastClipBeforeNew;
				if (lastClipBeforeNew == null)
					{
					AudioClip head = clipList.getFirst();
					if (head != null)
						{
						clip.next = head;
						head.prev = clip;
						}
					}
				else
					{
					clip.next = lastClipBeforeNew.next;
					if (lastClipBeforeNew.next != null)
						lastClipBeforeNew.next.prev = clip;
					lastClipBeforeNew.next = clip;
					}
				clipList.add(i, clip);
				}
			parentSong.updateLastBlock();
			}


		public AudioClip createClip(int id, AudioMedia audioSource, long trackInsertBlock, boolean isTake)
			{
			return createClip(id, audioSource, trackInsertBlock, 0, audioSource.totalBlocks, isTake);
			}


		public AudioClip createClip(int id, AudioMedia audioSource, long trackInsertBlock, long sourceFirstBlock, long lenght, boolean isTake)
			{
			log("Creating clip starting in " + trackInsertBlock);
			AudioClip clip = new AudioClip(id, this, audioSource, sourceFirstBlock, lenght, trackInsertBlock);
			addClip(clip);
			clip.isTake = isTake;
			return clip;
			}


		public void activate()
			{
			if (!isActive)
				{
				isActive = true;
				if (parentSong.editingMode == Const.EDITING_MODE_NORMAL)
					{
					int cfc = controllerChain.currentController;
						// MIG if ((cfc>=0) && (controllerChain.pluginController[cfc]))
						{
						// MIG controllerChain.pluginController[cfc].start_blinking();
						}
					}
				}
			}


		public void deactivate()
			{
			if (isActive)
				{
				for (int k = 0; k < Chain.MAX_CONTROLLERS; k++)
					{
					if (controllerChain.controller[k] != null)
						{
						controllerChain.controller[k].deactivate();
						}
					}
				isActive = false;
				}
			}


		public void setBlur(boolean stat)
			{
			setCmBlur(CM_TRACK_BG, stat);
			setCmBlur(CM_TRACK_BG_ACTIVE, stat);
			for (AudioClip c : clipList)
				{
				c.setBlur(stat);
				}
			}


		public int getBaseX() // the implementation for track is simply return 0
			{
			return 0;
			}


		public int getBaseY()
			{
			return baseY;
			}


		public int get_width() // the implementation for tracks is simply return the editor widths
			{
			return proxyEditor.getWidth();
			}


		public final int getHeight()
			{
			return height;
			}


		public String get_name()
			{
			return name;
			}


		public int arm()
			{
			/*
			 * MIG if ((mixer.get_status()==Mixer.STOPPED) || (mixer.get_status()==Mixer.ARMED)) { QDate d = QDate.currentDate(); QTime t = QTime.currentTime(); String s1;s1.setNum(d.year()); String s2;s2.setNum(d.month()); String s3;s3.setNum(d.day()); String s4;s4.setNum(t.hour()); String s5;s5.setNum(t.minute()); String s6;s6.setNum(t.second());
			 * String tns = "take-"+s1+s2+s3+"-"+s4+s5+s6+".praf"; recordingAudioFilename = tns; if (mixer.arm(busIn,whichChannels)<0) { //PERROR("Cannot arm track");//improve me return -1; } start_recording_clip(); isArmed=true; draw_panel_leds(); update_panel(); }
			 */
			return 1;
			}


		public int disarm()
			{
			/*
			 * MIG if ((mixer.get_status()==Mixer.STOPPED) || (mixer.get_status()==Mixer.ARMED)) // this avoid the disarm/arm during going state { isArmed=false; if (mixer.disarm(busIn)<0) { //PERROR("Cannot disarm track"); return -1; } abort_recording_clip(); draw_panel_leds(); recreate(); }
			 */
			return 1;
			}


		public int cleanup_buffers() // This is called everytime Mixer stops.
			{
			int r = 1;
			/*
			 * controllerChain.cleanup(); if (isArmed) { r=finish_recording_clip(); QDate d = QDate.currentDate(); QTime t = QTime.currentTime(); String s1;s1.setNum(d.year()); String s2;s2.setNum(d.month()); String s3;s3.setNum(d.day()); String s4;s4.setNum(t.hour()); String s5;s5.setNum(t.minute()); String s6;s6.setNum(t.second()); String tns =
			 * "take-"+s1+s2+s3+"-"+s4+s5+s6+".praf"; recordingAudioFilename = tns; start_recording_clip(); }
			 */
			return r;
			}


		public Song get_parent_song()
			{
			return parentSong;
			}


		public SoftBus getSoftIn()
			{
			return busIn;
			}


		public SoftBus getSoftOut()
			{
			return busOut;
			}


		public int get_which_channels()
			{
			return whichChannels;
			}


		public void set_which_channels(int wc)
			{
			whichChannels = wc; // TODO : validate it..
			}


		public void set_bus_in(SoftBus b)
			{
			boolean wasArmed = panel.ledArmed.state;
			if (panel.ledArmed.state)
				disarm();
			busIn = b;
			if (wasArmed)
				arm();
			}


		public void set_bus_out(SoftBus b)
			{
			busOut = b;
			}


		public boolean is_muted()
			{
			return panel.ledMute.state;
			}


		public boolean isSolo()
			{
			return panel.ledSolo.state;
			}


		public boolean armed()
			{
			return panel.ledArmed.state;
			}


		public boolean is_active()
			{
			return isActive;
			}


		public void toggle_active()
			{
			isActive = !isActive;
			}


		public String getSchema()
			{
			StringBuffer output = new StringBuffer("");
			output.append("<track id=\"" + trackNumber + "\"");
			output.append(" name=\"" + name + "\"");
			output.append(" gain=" + gain + " pan=" + pan);
			output.append(" mute=" + (panel.ledMute.state ? "1" : "0") + " solo=" + (panel.ledSolo.state ? "1" : "0"));
			output.append(" busIn=" + busIn.getId() + " busOut=" + busOut.getId());
			output.append(">\n");
			String tag;
			for (AudioClip cli : clipList)
				{
				tag = cli.isTake ? "take" : "source";
				output.append("       <clip " + tag + "=\"" + cli.audioSource.getPureName() + "\"");
				output.append(" id=" + cli.getId());
				output.append(" name=\"" + cli.getName() + "\"");
				output.append(" TFB=" + cli.trackFirstBlock);
				output.append(" SFB=" + cli.sourceFirstBlock);
				output.append(" LEN=" + cli.getLenght());
				output.append(" gain=" + cli.gain);
				output.append(" fadeIn=" + cli.fadeInBlocks);
				output.append(" fadeOut=" + cli.fadeOuBlocks + ">\n</clip>\n");
				}
			output.append(controllerChain.get_schema());
			output.append("</track>\n");
			return output.toString();
			}


		public void setSchema(String trackSchema)
			{
			name = SchemaParser.getXmlProperty(trackSchema, "name");
			panel.ledMute.state = SchemaParser.getXmlBooleanProperty(trackSchema, "mute");
			panel.ledSolo.state = SchemaParser.getXmlBooleanProperty(trackSchema, "solo");
			// UNCO set_gain(SchemaParser.getXmlFloatProperty(trackSchema, "gain"));
			// UNCO set_pan(SchemaParser.getXmlFloatProperty(trackSchema, "pan"));
			// MIG set_bus_in(SchemaParser.getXmlIntProperty(trackSchema, "busIn"));
			// MIG set_bus_out(SchemaParser.getXmlIntProperty(trackSchema, "busOut"));
			ArrayList<String> clipsSchema = SchemaParser.getSections(trackSchema, "clip");
			int t = clipsSchema.size();
			PMESG(4, "    Track " + get_id() + " : " + t + " clips");
			for (int i = 0; i < t; i++)
				{
				String clipSchema = clipsSchema.get(i);
				String pureSourceName = SchemaParser.getXmlProperty(clipSchema, "source");
				String clipName = SchemaParser.getXmlProperty(clipSchema, "name");
				int clipId = SchemaParser.getXmlIntProperty(clipSchema, "id");
				boolean isTake = SchemaParser.getXmlBooleanProperty(clipSchema, "isTake");
				long TFB = SchemaParser.getXmlLongProperty(clipSchema, "TFB");
				long SFB = SchemaParser.getXmlLongProperty(clipSchema, "SFB");
				long LEN = SchemaParser.getXmlLongProperty(clipSchema, "LEN");
				float clipGain = SchemaParser.getXmlFloatProperty(clipSchema, "gain");
				long fadeIn = SchemaParser.getXmlLongProperty(clipSchema, "fadeIn");
				long fadeOut = SchemaParser.getXmlLongProperty(clipSchema, "fadeOut");
				AudioClip clip = null;
				String phisicalSourceFilename = parentSong.rootDir + (isTake ? "/capture" : "/audiosources") + "/" + pureSourceName + "." + Const.PRAF_EXTENSION;
				PMESG(2, "Creating clip from " + phisicalSourceFilename + " TFB=" + TFB + " SFB=" + SFB + " LEN=" + LEN);
				AudioMedia audioSource = getAudioSource(pureSourceName);
				if (audioSource == null)
					{
					try
						{
						audioSource = parentSong.audioSourcesList.loadSource(phisicalSourceFilename);
						}
					catch (IOException ex)
						{
						PERROR("Audio source " + pureSourceName + " specified for track " + get_id() + " could not be loaded. Ignoring it... ");
						ex.printStackTrace();
						}
					}
				if (audioSource != null)
					{
					clip = createClip(clipId, audioSource, TFB, SFB, LEN, isTake);
					clip.setGain(clipGain);
					clip.setFadeIn(fadeIn);
					clip.setFadeOut(fadeOut);
					clip.setName(clipName);
					}
				}
			/*
			 * else if (line.find("<filtercontroller")>=0) { int x1=line.find("type=")+6; int x2=line.find("\"",x1); String fctype=line.mid(x1,x2-x1); AudioPluginController* fc = controllerChain.add_audio_plugin_controller(fctype); if (fc) { while ( fc && (line.find("</filtercontroller")<0) && (!ts.atEnd())) { line = ts.readLine(); line =
			 * line.stripWhiteSpace(); if (line.find("<curve")>=0) { int x1=line.find("type=")+6; int x2=line.find("\"",x1); String cutype=line.mid(x1,x2-x1); //PMESG("Adding Curve %s",cutype); Curve* cu = fc.add_curve(cutype); if (!cu) // curve already exists. No need to re-add it. cu = fc.get_curve(cutype); // use the already existent curve
			 * while (cu && (line.find("</curve")<0) && (!ts.atEnd())) { line = ts.readLine(); line = line.stripWhiteSpace(); if (line.find("<node")>=0) { int x1=line.find("pos=")+4; int x2=line.find(" ",x1); String spos=line.mid(x1,x2-x1); int x3=line.find("value=")+6; int x4=line.find(" ",x3); String sval=line.mid(x3,x4-x3); pos= 0;//MIG(long)
			 * spos.toDouble(); if (pos==0) { // //PMESG("Setting root node [%s@%s]",sval,spos); } else { ////PMESG("Adding node : [%s@%s]",sval,spos); } float val= sval.toFloat(); CurveNode* no=cu.add_node(pos,val); } } } } } } } }
			 */
			}


		public int start_recording_clip()
			{
			int recordingChannels;
			if (whichChannels == Const.ONLY_LEFT_CHANNEL || whichChannels == Const.ONLY_RIGHT_CHANNEL || whichChannels == Const.MONO)
				recordingChannels = 1;
			else
				recordingChannels = 2;
			// MIG recordingAudio = new Audio(parentSong,parentSong.captureDir, recordingAudioFilename, recordingChannels, parentSong.rate, parentSong.bitDepth, "Untitled", "UndefGroup", "UndefSoundCard", "Real Time Take");
			Peak peak = recordingAudio.getPeak();
			;
			try
				{
				peak.defineSourceMedia(recordingAudio);
				}
			catch (PeakException ex)
				{
				ex.printStackTrace();
				}
			// IMPROVE LINE BELOW. File is already isolated within Audio
			recordingAudioFile = recordingAudio;
			recordingAudioFile.rewindAudio();
			recordingAudioStartPos = parentSong.workingBlock;
			recordingAudioPos = recordingAudioStartPos;
			// MIG updateTimer.start(50);
			return 1; // improve me
			}


		public int abort_recording_clip()
			{
			String name = parentSong.captureDir + "/" + recordingAudioFilename;
			// MIG File file = new File(name);
			// MIG if (!file.remove()) //file is not used, remove it from hard disk
			// MIG {
			// MIG //PERROR("failed to remove file %s\n", name);
			// MIG }
			// delete recordingAudio;
			// MIG updateTimer.stop();
			return 0;
			}


		public void feed_recording_clip(byte[] buffer, int bufferSize)
			{
			recordingBufferSize = bufferSize;
			recordingBuffer = buffer;
			try
				{
				Peak peak = recordingAudio.getPeak();
				recordingAudioFile.writeFragment(recordingBuffer, recordingBufferSize);
				peak.process_samples_from_buffer(buffer, bufferSize);
				recordingAudioPos += recordingBufferSize / recordingAudioFile.blockSize;
				}
			catch (IOException e)
				{
				PERROR("ERROR :" + e);
				e.printStackTrace();
				}
			}


		public int finish_recording_clip()
			{
			// MIG updateTimer.stop();
			int r = accept_recorded_clip();
			return r;
			}


		public void update_recording_clip()
			{
			/*
			 * MIG if (mixer.get_status()==Mixer.GOING) { //PMESG3("Updating recording Audio %p",recordingAudio); int recordingChannels; if(whichChannels==AudioDeviceMapper.ONLY_LEFT_CHANNEL || whichChannels==AudioDeviceMapper.ONLY_RIGHT_CHANNEL || whichChannels==AudioDeviceMapper.MONO) recordingChannels = 1; else recordingChannels = 2; int rx =
			 * parentSong.block_to_xpos(recordingAudioPos) + proxyEditor.clipAreaBaseX; int totalSamples = recordingBufferSize/(2FIXME/(parentSong.bitDepth/8)); int numSamplesToDraw = totalSamples/Peak.zoomStep[parentSong.hzoom]; QPainter* painter = editor.painter; if (( rx >= 0 ) && (rx <= proxyEditor.clipAreaWidth)) { if ( rx + numSamplesToDraw >
			 * proxyEditor.clipAreaWidth ) numSamplesToDraw = proxyEditor.clipAreaWidth - rx; int recordingBufferStep = Peak.zoomStep[parentSong.hzoom]*(2FIXME*(parentSong.bitDepth/8)); int sho,maxsho; int chhalf = (height-12) / recordingChannels; int x = rx - 2; // this 2 pixels is to prevent that the recording buffer being drawed is blurred by
			 * playing cursors, that is being drawed in the same position during record int sampleCounter=0; if (numSamplesToDraw==0) { recordingBufferStep = recordingBufferSize; numSamplesToDraw=1; } char* posBuffer = recordingBuffer; while ( sampleCounter < numSamplesToDraw ) { for (int k=0; k < recordingChannels; k++) { int yc = real_baseY() +
			 * chhalf*(k+1) + 4 ; painter.setColor(Color.white); painter.drawLine(x,yc,x,yc-chhalf+1); char* p = posBuffer + k*sizeof(short); maxsho=0; while ( p < (posBuffer + recordingBufferStep)) { sho = (int)*(short*)p; if ( sho > maxsho ) maxsho=sho; p += sizeof(short)*recordingChannels; } if ( maxsho >= 32767 ) { maxsho = 32767;
			 * painter.setColor(Color.red); inputLevel[k]=maxsho; } else painter.setColor(new Color(150,40,40)); float f = (float) maxsho / 32768; int y = (int) (f*chhalf); painter.drawLine(x,yc,x,yc-y); } posBuffer+=recordingBufferStep; x++; sampleCounter++; } } editor.update((rx-2),real_baseY(),(rx + numSamplesToDraw -2),height); }
			 */
			}


		public int accept_recorded_clip()
			{
			try
				{
				recordingAudioFile.fixAudioSize();
				recordingAudio.buildPeaks(); // WAS: rebuild_peaks(); As this is recorded audio, it's impossible that a peakfile allready exist, so just build it, not rebuild (rebuild removes the existing Peak Object, which make's it impossible to generate a peakfile on the fly. By Remon
				}
			catch (Exception ex)
				{
				System.out.println("Error : " + ex);
				}
			AudioClip clip = createClip(AudioClip.lastId++, recordingAudio, recordingAudioStartPos, true);
			if (clip == null)
				{
				// PERROR("After recording, could not create clip with recorded audio! Continuing anyway ...");
				return -1;
				}
			addClip(clip);
			lastRecordedClip = clip;
			// MIG parentSong.set_work_at(lastRecordedClip.trackLastBlock);
			// MIG parentSong.audioSourcesList.add(recordingAudio);
			return 1;
			}


		public void discard_recorded_clip()
			{
			// delete recordingAudio;
			}


		public float getGain()
			{
			return gain;
			}


		public float getPan()
			{
			return pan;
			}


		public void set_gain(float pGain)
			{
			if (pGain < -24.0)
				gain = -24.0f;
			else if (pGain > 24.0)
				gain = 24.0f;
			else
				gain = pGain;
			// FOR FUTURE update_panel();
			}


		public void set_pan(float pPan)
			{
			if (pPan < -1.0)
				pan = -1.0f;
			else if (pPan > 1.0)
				pan = 1.0f;
			else
				pan = pPan;
			// FOR FUTURE update_panel();
			}


		public void vzoom_in(int newBaseY)
			{
			baseY = newBaseY;
			if (1.1 * height < 500.0)
				height = (int) (1.10 * height);
			}


		public void vzoom_out(int newBaseY)
			{
			baseY = newBaseY;
			if (0.9 * height > Const.MINIMUM_TRACK_HEIGHT)
				height = (int) (0.90 * height);
			}


		public void set_baseY(int b)
			{
			baseY = b;
			}


		public void set_height(int h)
			{
			height = h;
			}


		public int get_total_clips()
			{
			return clipList.size();
			}


		public boolean visible()
			{
			int th = proxyEditor.getHeight();
			int my = Const.EDITOR_AND_RACK_TRACKS_BASE_Y;
			if ((baseY + height + my) < 0)
				return false;
			if ((baseY + my) > th)
				return false;
			return true;
			}


		public int getRealBaseY() // return baseY with vertical offset applied
			{
			return baseY + Const.EDITOR_AND_RACK_TRACKS_BASE_Y;
			}


		public int get_mouse_x()
			{
			return proxyEditor.getMouseX();
			}


		public int get_mouse_y()
			{
			return proxyEditor.getMouseY();
			}


		public void toggle_solo()
			{
			panel.ledSolo.toggle();
			}


		public void toggle_mute()
			{
			panel.ledMute.toggle();
			}


		protected AudioMedia getAudioSource(String sourceName)
			{
			return parentSong.audioSourcesList.get_audio_for_source(sourceName);
			}


		void resize_panel_pixmap()
			{
			// FOR FUTURE panelPixmap.resize(proxyEditor.trackPanelAreaWidth, 200);
			}


		public boolean someClipPeaksBecameAvailable()
			{
			for (AudioClip c : clipList)
				{
				if (c.isVisible() && !c.isWaitingPeaks(parentSong.hzoom) && (!c.peaksDrawn))
					{
					// log("Clip "+c.clipName+" got their peaks available in track " + get_id() + ". Recreating it.");
					return true;
					}
				}
			return false;
			}


		protected final void log(String message)
			{
			PMESG(4, "[ Track " + get_id() + " ] " + message);
			}


		public boolean someClipsWaitingForPeaks()
			{
			int hz = parentSong.hzoom;
			for (AudioClip c : clipList)
				{
				if (c.isVisible() && c.isWaitingPeaks(hz))
					{
					c.draw();
					return true;
					}
				}
			return false;
			}


		public boolean someClipsBlocked()
			{
			int hz = parentSong.hzoom;
			for (AudioClip c : clipList)
				{
				if (c.isVisible() && c.peaksBlocked(hz))
					{
					return true;
					}
				}
			return false;
			}


		/*
		 * FOR FUTURE public void update_panel() { int panelWidth = proxyEditor.trackPanelAreaWidth; // FOR FUTURE (draw cached panel) // painter.drawImage(panelPixmap, 0, 0, panelWidth, height); // MEANWHILE proxyEditor.update(0, real_baseY(), panelWidth, height); }
		 */

		public boolean isMuted()
			{
			return panel.ledMute.state;
			}


		public boolean isActive()
			{
			return isActive;
			}


		/**
		 * Draw just a countour of the passed clip, in order to preview how it will be placed after a drag, a copy, or some like that...
		 * @param clip
		 */
		public void drawGhostClip(Graphics2D painter, AudioClip clip)
			{
			recreateClipArea();
			int bx = parentSong.blockToEditorX(clip.trackFirstBlock);
			int by = getRealBaseY();
			int w = parentSong.blockToEditorX(clip.trackLastBlock) - bx;
			int h = getHeight();
			painter.drawRect(bx, by, w, h - 2);
			}


	}
