package org.protux.core.gui;

import static org.protux.core.Debugger.PMESG;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JPanel;

/**
 * Whoever extends FastPanel must follow these conventions :
 * <ul>
 * 		<li>draw clients (things that calls drawXXX() methods) must repaint , preferrentially only the portion drawn
 * 		<li>recreate methods are draw clients , so have the responsability to repaint
 * 		<li>draw methods must not repaint
 * 		<li>draw methods must not clear
 * 		<li>clear methos must not repaint
 * 		<li>clear methods must not draw
 * </ul>
 */

public abstract class FastPanel extends JPanel implements MouseListener
	{
		private static final Dimension STANDARD_MIN_SIZE = new Dimension(10, 10);
		public boolean logging = false;
		public String panelName;
		public Dimension lastSize = null;

		private Image softwareBufferedImage;


		public FastPanel(String panelName, Dimension dim, Color bgColor)
			{
			super(true);
			this.panelName = panelName;
			this.setOpaque(true);
			this.setIgnoreRepaint(true);
			this.setBackground(bgColor);
			this.setMinimumSize(dim);
			this.setPreferredSize(dim);
			this.softwareBufferedImage = createImage(getWidth(), getHeight());
			this.addMouseListener(this);

			}


		public FastPanel(String panelName)
			{
			this(panelName, STANDARD_MIN_SIZE, Color.BLACK);
			}


		@Override
		public void paintComponent(Graphics painter)
			{
			// super.paintComponent(painter);
			Dimension currentSize = getSize();
			if ((lastSize == null) || (!lastSize.equals(currentSize)))
				{
				lastSize = currentSize;
				if (logging)
					log("Resizing " + panelName + " to " + lastSize.width + "x" + lastSize.height + "( Invoking recreation )");
				softwareBufferedImage = createImage(currentSize.width, currentSize.height);
				recreate(getGBuffer());
				}
			painter.drawImage(softwareBufferedImage, 0, 0, this);
			painter.dispose();
			}


		public void recreate()
			{
			Graphics2D painter = getGBuffer();
			if (painter != null)
				{
				recreate(painter);
				repaint();
				}
			}


		public abstract void recreate(Graphics2D painter);


		public final Graphics2D getGBuffer()
			{
			if (softwareBufferedImage != null)
				return (Graphics2D) softwareBufferedImage.getGraphics();
			return null;
			}


		@Override
		public final void repaint()
			{
			Graphics g = super.getGraphics();
			repaint(g);
			}


		public final void repaint(Graphics g)
			{
			if ((g != null) && (softwareBufferedImage != null))
				{
				System.out.println("PAINTING FULL " + panelName);
				g.drawImage(softwareBufferedImage, 0, 0, this);
				}
			}


		/**
		 * Same as repaint(x,y,w,h), but uses a fastcanvas region coordinates instead. Available as a convenience
		 */
		public final void repaint(Rectangle region)
			{
			repaint(region.baseX, region.baseY, region.width, region.height);
			}


		@Override
		public final void repaint(int x, int y, int w, int h)
			{
			repaint(super.getGraphics(), x, y, w, h);
			}


		public final void repaint(Graphics g, int x, int y, int w, int h)
			{
			g.drawImage(softwareBufferedImage, x, y, x + w, y + h, x, y, x + w, y + h, this);
			}


		@Override
		@Deprecated
		/**
		 * Deprecated purposefully to inhibit clients to access it.
		 */
		public Graphics getGraphics()
			{
			return super.getGraphics();
			}



		// -------- Convenience methods ----------

		protected void drawDumbX(Graphics2D painter)
			{
			if (painter != null)
				{
				painter.setColor(Color.RED);
				painter.drawLine(0, 0, getWidth(), getHeight());
				painter.drawLine(getWidth(), 0, 0, getHeight());
				}
			}

		public final void clear()
			{
			clear(getGBuffer());
			}


		public final void clear(Graphics2D painter)
			{
			painter.clearRect(0, 0, getWidth(), getHeight());
			}


		public final void clear(Graphics2D painter, int x, int y, int w, int h)
			{
			painter.clearRect(x, y, w, h);
			}

		private final void log(String string)
			{
			PMESG(2, "[ FastCanvas " + panelName + " ] " + string);
			}


		
	}
