package org.protux.core.gui;

import static org.protux.core.Debugger.PMESG;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;
import java.util.TimerTask;

import org.protux.core.OneShotTimer;

public class Lcd extends FastPanel implements InfoViewer
	{
		protected String welcomeText;
		protected int maxDigits;
		protected int welcomeIndex;
		protected int rollLine = 0;
		protected static final int DOT_SIZE = 2;
		protected static final int gridDotSize = 1;
		protected static final int digitHSpace = 1;
		protected static final int digitVSpace = 1;
		protected static final int digitHSize = 5;
		protected static final int digitVSize = 8;
		protected static final int dotVSpace = 2;
		protected static final int dotHSpace = 2;
		protected static final int LCD_HORIZONTAL_MARGIN = 8;
		protected static final int LCD_VERTICAL_MARGIN = 4;
		protected static final int SCALE = 2;
		protected static final int TOTAL_DIGIT_WIDTH = (digitHSize + digitHSpace) * SCALE;
		protected static final int TOTAL_LINE_HEIGHT = (digitVSize + digitVSpace) * SCALE;
		protected static final int NUM_LCD_LINES = 2;

		private static final Color COLOR_BG = new Color(220, 235, 201);
		private static final Color COLOR_DIGIT_GRID = new Color(180, 196, 172);
		private static final Color COLOR_DIGIT_DOT = new Color(20, 20, 40);
		private static final int MIN_LCD_WIDTH = 100;

		protected OneShotTimer welcomeTimer = new OneShotTimer();


		public Lcd()
			{
			super("LCD", new Dimension(MIN_LCD_WIDTH, NUM_LCD_LINES * TOTAL_LINE_HEIGHT + LCD_VERTICAL_MARGIN * 2), COLOR_BG);
			refreshMaxDigits();
			}


		private void refreshMaxDigits()
			{
			maxDigits = 25; // FIXME it must calculate the number of digits depending on the lcd width.
			}


		public void rollMessage(String text, int vel, int line)
			{
			welcomeText = text;
			welcomeIndex = 0;
			welcomeTimer.schedule(new WelcomeTask(), 2000, vel);
			rollLine = line;
			}


		public int getMaxDigits()
			{
			return maxDigits;
			}


		@Override
		public void recreate(Graphics2D painter)
			{
			log("Recreating...");
			clear(painter);
			drawAllGrids(painter);
			}


		private void clearDigit(Graphics2D g2, int line, int col)
			{
			int x = getBaseY(col);
			int y = getBaseY(line);
			g2.setColor(COLOR_BG);
			g2.fillRect(x, y, TOTAL_DIGIT_WIDTH, TOTAL_LINE_HEIGHT);
			drawDigitGrid(g2, line, col);
			}


		private void clearDigits(Graphics2D g2, int line, int startCol, int endCol)
			{
			for (int i = startCol; i <= endCol; i++)
				{
				clearDigit(g2, line, i);
				}
			}


		protected void clearLine(Graphics2D g2, int line)
			{
			clearDigits(g2, line, 0, maxDigits);
			}


		private void drawDigitGrid(Graphics2D g2, int line, int col)
			{
			g2.setColor(COLOR_DIGIT_GRID);
			for (int k = 0; k < digitVSize; k++)
				for (int j = 0; j < digitHSize; j++)
					{
					int x = LCD_HORIZONTAL_MARGIN + col * TOTAL_DIGIT_WIDTH + (j * SCALE);
					int y = LCD_VERTICAL_MARGIN + line * TOTAL_LINE_HEIGHT + (k * SCALE);
					g2.drawRect(x, y, gridDotSize, gridDotSize);
					}
			}


		private void drawDigitsGrids(Graphics2D g2, int line, int startCol, int endCol)
			{
			for (int c = startCol; c <= endCol; c++)
				{
				drawDigitGrid(g2, line, c);
				}
			}


		private void drawLineGrid(Graphics2D g2, int line)
			{
			drawDigitsGrids(g2, line, 0, maxDigits);
			}


		protected void drawAllGrids()
			{
			Graphics2D g = getGBuffer();
			if (g != null)
				{
				drawAllGrids(g);
				}
			}


		protected void drawAllGrids(Graphics2D g2)
			{
			for (int i = 0; i < NUM_LCD_LINES; i++)
				{
				drawLineGrid(g2, i);
				}
			}


		private void drawDigit(Graphics2D g2, int index, int line, int col)
			{
			g2.setColor(COLOR_DIGIT_DOT);
			final int by = getBaseY(line);
			for (int x = 0; x < digitHSize; x++)
				{
				for (int y = 0; y < digitVSize; y++)
					{
					char c = digit[index][y].charAt(x);
					if (c == 'O')
						{
						g2.fillRect(LCD_HORIZONTAL_MARGIN + col * TOTAL_DIGIT_WIDTH + (x * SCALE), by + (y * SCALE), DOT_SIZE, DOT_SIZE);
						}
					}
				}
			}


		protected final int getBaseY(int line)
			{
			return LCD_VERTICAL_MARGIN + (line * TOTAL_LINE_HEIGHT);
			}


		protected final int getBaseX(int col)
			{
			return LCD_HORIZONTAL_MARGIN + (col * TOTAL_DIGIT_WIDTH);
			}


		private void print(Graphics2D g2, String s, int line)
			{
			print(g2, s, line, 0, true);
			}


		protected void print(Graphics2D g2, String s, int line, int col, boolean clean)
			{
			if (g2 == null) // necesary because sometimes print is called within a non visible lcd
				{
				return;
				}
			if (clean)
				{
				clearLine(g2, line);
				}
			else
				{
				clearDigits(g2, line, col, col + s.length());
				}
			int i = 0;
			while (i < s.length())
				{
				char cs = s.charAt(i);
				drawDigit(g2, cs - 32, line, i + col);
				i++;
				}
			}


		private void draw_background_grid(Graphics2D g2)
			{
			int gridxpoints = (getWidth() - 10) / dotHSpace;
			int gridypoints = (getHeight() - 10) / dotVSpace;
			for (int i = 0; i < gridxpoints; i++)
				for (int j = 0; j < gridypoints; j++)
					{
					g2.fillRect(5 + i * dotHSpace, 5 + j * dotVSpace, 1, 1);
					}
			}


		public void displayStatus(String s)
			{
			Graphics2D g2 = getGBuffer();
			print(g2, s, 2);
			}


		public void displayMainMessage(String s)
			{
			Graphics2D g2 = getGBuffer();
			print(g2, s, 1);
			//repaint();
			}


		public void displayInstruction(String s)
			{
			Graphics2D g2 = getGBuffer();
			print(g2, s, 3);
			//repaint();
			}


		public void displayDirections(boolean up, boolean down, boolean left, boolean right)
			{
			if (up)
				{

				}
			if (down)
				{

				}
			if (left)
				{

				}
			if (right)
				{

				}
			//repaint();
			}


		public void displayError(String s)
			{
			Graphics2D g2 = getGBuffer();
			print(g2, s, 1);
			//repaint();
			}

		private class WelcomeTask extends TimerTask
			{
				private boolean started = false;


				@Override
				public void run()
					{
					if (!started)
						{
						PMESG(3, "Rolling Message: " + welcomeText);
						}
					started = true;
					Graphics2D g2 = getGBuffer();
					int by = getBaseY(rollLine);
					String s = "";
					int md = getMaxDigits();
					if (welcomeIndex < (md + 20))
						{
						s = welcomeText.substring(welcomeIndex, welcomeIndex + 1);
						if ((!s.equals("")) && (welcomeIndex < md))
							{
							print(g2, s, rollLine, welcomeIndex, false);
							}
						}
					else
						{
						clearLine(g2, rollLine);
						s = welcomeText.substring(welcomeIndex - 20 - md, welcomeIndex - 20);
						if (!s.equals(""))
							{
							print(g2, s, rollLine, 0, false);
							}
						}
					synchronized (this)
						{
						repaint(0, by, getWidth(), by + TOTAL_LINE_HEIGHT);
						}
					welcomeIndex++;
					if (welcomeIndex >= welcomeText.length() + 20)
						{
						welcomeTimer.stop();
						}
					}
			}


		private void log(String string)
			{
			PMESG(2, "[ Lcd ] " + string);
			}

		private static final String digit[][] = {
				{ "     ", "     ", "     ", "     ", "     ", "     ", "     ", "     " }, // space
				{ "  O  ", "  O  ", "  O  ", "  O  ", "  O  ", "     ", "  O  ", "     " }, // !
				{ "     ", "     ", " O O ", " O O ", "     ", "     ", "     ", "     " }, // "
				{ " O O ", " O   ", "OOOOO", " O O ", "OOOOO", " O O ", " O O ", "     " }, // #
				{ "  O  ", " OOO ", "O   O", " OO  ", "  OO ", "    O", " OOO ", "  O  " }, // $
				{ " O   ", "O O O", " O O ", "  O  ", " O O ", "O O O", "   O ", "     " }, // %
				{ " OOO ", "O   O", " O   ", " OO  ", "O O  ", "O  O ", " OO O", "     " }, // &
				{ "  O  ", "  O  ", "     ", "     ", "     ", "     ", "     ", "     " }, // '
				{ "   O ", "  O  ", " O   ", " O   ", " O   ", " O   ", "  O  ", "   O " }, // (
				{ " O   ", "  O  ", "   O ", "   O ", "   O ", "   O ", "  O  ", " O   " }, // )
				{ "  O  ", "O O O", " OOO ", "OOOOO", " OOO ", "O O O", "  O  ", "     " }, // *
				{ "     ", "  O  ", "  O  ", "OOOOO", "  O  ", "  O  ", "     ", "     " }, // +
				{ "     ", "     ", "     ", "     ", "     ", "     ", "  O  ", " O   " }, // ,
				{ "     ", "     ", "     ", " OOO ", "     ", "     ", "     ", "     " }, // -
				{ "     ", "     ", "     ", "     ", "     ", "     ", "  O  ", "     " }, // .
				{ "   O ", "   O ", "  O  ", "  O  ", " O   ", " O   ", "     ", "     " }, // /
				{ "  O  ", " O O ", "O   O", "O   O", "O   O", " O O ", "  O  ", "     " }, //
				{ "  O  ", " OO  ", "  O  ", "  O  ", "  O  ", "  O  ", " OOO ", "     " }, // O
				{ " OOO ", "O   O", "    O", "  OO ", " O   ", "O    ", "OOOOO", "     " }, // 2
				{ " OOO ", "O   O", "    O", "  OO ", "    O", "O   O", " OOO ", "     " }, // 3
				{ "  OO ", " O O ", "O  O ", "OOOO ", "   O ", "   O ", "  OOO", "     " }, // 4
				{ "OOOOO", "O    ", "O    ", "OOOO ", "    O", "O   O", " OOO ", "     " }, // 5
				{ " OOO ", "O   O", "O    ", "OOOO ", "O   O", "O   O", " OOO ", "     " }, // 6
				{ "OOOOO", "    O", "   O ", "  O  ", "  O  ", "  O  ", "  O  ", "     " }, // 7
				{ " OOO ", "O   O", "O   O", " OOO ", "O   O", "O   O", " OOO ", "     " }, // 8
				{ " OOO ", "O   O", "O   O", " OOOO", "    O", "    O", " OOO ", "     " }, // 9
				{ "     ", "     ", "  O  ", "     ", "  O  ", "     ", "     ", "     " }, // :
				{ "     ", "     ", "     ", "  O  ", "     ", "  O  ", "  O  ", " O   " }, // ;
				{ "   O ", "  O  ", " O   ", "O    ", " O   ", "  O  ", "   O ", "     " }, // <
				{ "     ", "     ", " OOO ", "     ", " OOO ", "     ", "     ", "     " }, // =
				{ " O   ", "  O  ", "   O ", "    O", "   O ", "  O  ", " O   ", "     " }, // >
				{ " OOO ", "O   O", "    O", "   O ", "  O  ", "     ", "  O  ", "     " }, // ?
				{ " OOOO", "O   O", "O  OO", "O O O", "O  OO", "O    ", " OOO ", "     " }, // @
				{ "  O  ", " O O ", "O   O", "O   O", "OOOOO", "O   O", "O   O", "     " }, // A
				{ "OOOO ", "O   O", "O   O", "OOOO ", "O   O", "O   O", "OOOO ", "     " }, // B
				{ " OOO ", "O   O", "O    ", "O    ", "O    ", "O   O", " OOO ", "     " }, // C
				{ "OOOO ", "O   O", "O   O", "O   O", "O   O", "O   O", "OOOO ", "     " }, // D
				{ "OOOOO", "O    ", "O    ", "OOOO ", "O    ", "O    ", "OOOOO", "     " }, // E
				{ "OOOOO", "O    ", "O    ", "OOOO ", "O    ", "O    ", "O    ", "     " }, // F
				{ " OOO ", "O   O", "O    ", "O  OO", "O   O", "O   O", " OOO ", "     " }, // G
				{ "O   O", "O   O", "O   O", "OOOOO", "O   O", "O   O", "O   O", "     " }, // H
				{ " OOO ", "  O  ", "  O  ", "  O  ", "  O  ", "  O  ", " OOO ", "     " }, // I
				{ "  OOO", "   O ", "   O ", "   O ", "O  O ", "O  O ", " OO  ", "     " }, // J
				{ "O   O", "O  O ", "O O  ", "OO   ", "O O  ", "O  O ", "O   O", "     " }, // K
				{ "O    ", "O    ", "O    ", "O    ", "O    ", "O    ", "OOOOO", "     " }, // L
				{ "O   O", "OO OO", "O O O", "O O O", "O   O", "O   O", "O   O", "     " }, // M
				{ "O   O", "OO  O", "O O O", "O O O", "O  OO", "O   O", "O   O", "     " }, // N
				{ " OOO ", "O   O", "O   O", "O   O", "O   O", "O   O", " OOO ", "     " }, // O
				{ "OOOO ", "O   O", "O   O", "OOOO ", "O    ", "O    ", "O    ", "     " }, // P
				{ " OOO ", "O   O", "O   O", "O   O", "O   O", "O  O ", " OO O", "     " }, // Q
				{ "OOOO ", "O   O", "O   O", "OOOO ", "O O  ", "O  O ", "O   O", "     " }, // R
				{ " OOO ", "O   O", "O    ", " OOO ", "    O", "O   O", " OOO ", "     " }, // S
				{ "OOOOO", "  O  ", "  O  ", "  O  ", "  O  ", "  O  ", "  O  ", "     " }, // T
				{ "O   O", "O   O", "O   O", "O   O", "O   O", "O   O", " OOO ", "     " }, // U
				{ "O   O", "O   O", "O   O", "O   O", "O   O", " O O ", "  O  ", "     " }, // V
				{ "O   O", "O   O", "O   O", "O O O", "O O O", " O O ", " O O ", "     " }, // W
				{ "O   O", "O   O", " O O ", "  O  ", " O O ", "O   O", "O   O", "     " }, // X
				{ "O   O", "O   O", " O O ", "  O  ", "  O  ", "  O  ", "  O  ", "     " }, // Y
				{ "OOOOO", "    O", "   O ", "  O  ", " O   ", "O    ", "OOOOO", "     " }, // Z
				{ "     ", "     ", "     ", "     ", "     ", "     ", "     ", "     " }, { "     ", "     ", "     ", "     ", "     ", "     ", "     ", "     " }, { "     ", "     ", "     ", "     ", "     ", "     ", "     ", "     " }, { "     ", "     ", "     ", "     ", "     ", "     ", "     ", "     " },
				{ "     ", "     ", "     ", "     ", "     ", "     ", "     ", "     " }, { "     ", "     ", "     ", "     ", "     ", "     ", "     ", "     " }, { "     ", "     ", " OOO ", "    O", " OOOO", "O   O", " OOOO", "     " }, // a
				{ "O    ", "O    ", "OOOO ", "O   O", "O   O", "O   O", "OOOO ", "     " }, // b
				{ "     ", "     ", " OOO ", "O   O", "O    ", "O   O", " OOO ", "     " }, // c
				{ "    O", "    O", " OOOO", "O   O", "O   O", "O   O", " OOOO", "     " }, // d
				{ "     ", "     ", " OOO ", "O   O", "OOOOO", "O    ", " OOO ", "     " }, // e
				{ "   OO", "  O  ", "  O  ", "  O  ", " OOO ", "  O  ", "  O  ", "     " }, // f
				{ "     ", "     ", " OOO ", "O   O", "O   O", " OOOO", "    O", " OOO " }, // g
				{ "O    ", "O    ", "OOOO ", "O   O", "O   O", "O   O", "O   O", "     " }, // h
				{ "  O  ", "     ", " OO  ", "  O  ", "  O  ", "  O  ", "  O  ", "     " }, // i
				{ "   O ", "     ", "  OOO", "   O ", "   O ", "   O ", "O  O ", " OO  " }, // j
				{ "     ", "O    ", "O O  ", "OO   ", "OO   ", "O O  ", "O  O ", "     " }, // k
				{ " OO  ", "  O  ", "  O  ", "  O  ", "  O  ", "  O  ", " OOO ", "     " }, // l
				{ "     ", "     ", "OO O ", "O O O", "O O O", "O O O", "O O O", "     " }, // m
				{ "     ", "     ", "O OO ", " O  O", " O  O", " O  O", " O  O", "     " }, // n
				{ "     ", "     ", " OOO ", "O   O", "O   O", "O   O", " OOO ", "     " }, // o
				{ "     ", "     ", "O OO ", " O  O", " O  O", " O  O", " OOO ", " O   " }, // p
				{ "     ", "     ", " OO O", "O  O ", "O  O ", "O  O ", " OOO ", "   O " }, // q
				{ "     ", "     ", "O OO ", " O  O", " O   ", " O   ", " O   ", "     " }, // r
				{ "     ", "     ", " OOOO", "O    ", " OOO ", "    O", "OOOO ", "     " }, // s
				{ "     ", " O   ", "OOOO ", " O   ", " O   ", " O   ", "  OO ", "     " }, // t
				{ "     ", "     ", "O   O", "O   O", "O   O", "O   O", " OOO ", "     " }, // u
				{ "     ", "     ", "O   O", "O   O", "O   O", " O O ", "  O  ", "     " }, // v
				{ "     ", "     ", "O   O", "O O O", "O O O", "O O O", " OOO ", "     " }, // w
				{ "     ", "     ", "O   O", " O O ", "  O  ", " O O ", "O   O", "     " }, // x
				{ "     ", "     ", "O   O", "O   O", "O   O", " OOOO", "    O", " OOO " }, // y
				{ "     ", "     ", "OOOOO", "   O ", "  O  ", " O   ", "OOOOO", "     " } // z
		};


		public void mouseClicked(MouseEvent e)
			{
			}


		public void mousePressed(MouseEvent e)
			{
			}


		public void mouseReleased(MouseEvent e)
			{
			}


		public void mouseEntered(MouseEvent e)
			{
			}


		public void mouseExited(MouseEvent e)
			{
			}
	}
