package org.protux.core.media.audio;

import static org.protux.core.Debugger.PERROR;
import static org.protux.core.Debugger.PMESG;
import static org.protux.core.Debugger.PWARN;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.channels.FileChannel;

import org.protux.core.Byter;
import org.protux.core.Const;
import org.protux.core.Util;
import org.protux.core.editor.Peak;
import org.protux.core.exceptions.PeakException;

public abstract class AudioMedia
	{
		protected RandomAccessFile raFile;
		protected char formatLabel[] = new char[4];

		// using short[] for header (instead of byte[]) to emulate unsigned byte, since they dont exist in Java ..
		// see http://darksleep.com/player/JavaAndUnsignedTypes.html#why_no_unsigned_types
		// to see a explanation for this shame...
		protected short header[] = null;

		public int channels = 0;
		public int rate = 0;
		public int bitDepth = 0;
		public int blockSize = 0;
		public int headerSize = 0;
		public int bytesPerSec = 0;
		public int sampleSize = 0;
		public long audioSize = 0;
		public long totalBlocks = 0;
		public File phisicalFile = null;
		public String projectLabel = "";
		public String groupLabel = "";
		public String hardwareLabel = "";
		public String description = "";
		protected Peak peak;
		protected boolean peakLess = false;
		protected int bestFragSize = 4096;
		protected ByteBuffer byteBufferForReading = null;
		private int numSamplesToRead = -1;
		private FileChannel fileChannel;
		private boolean clean32BitsSupport = false;


		public AudioMedia(File aFile, int headerSize) throws IOException
			{
			this(aFile, false, headerSize, false);
			}


		public AudioMedia(File aFile, boolean peakLess, int headerSize) throws IOException
			{
			this(aFile, peakLess, headerSize, false);
			}


		public AudioMedia(File aFile, boolean peakLess, int headerSize, int mode) throws IOException
			{
			this(aFile, peakLess, headerSize, false || (mode == Const.READ_ONLY));
			}


		public AudioMedia(File aFile, boolean peakLess, int headerSize, boolean ro) throws IOException
			{
			this.phisicalFile = aFile;
			this.headerSize = headerSize;
			this.peakLess = peakLess;
			header = new short[headerSize];
			if (ro)
				{
				raFile = new RandomAccessFile(aFile, "r");
				}
			else
				{
				raFile = new RandomAccessFile(aFile, "rw");
				}
			if (peakLess)
				peak = null;
			else
				peak = new Peak();
			fileChannel = raFile.getChannel();
			}


		public void set(int rate, int channels, int bitDepth, String projLab, String groupLab, String hardLab, String descLab)
			{
			this.rate = rate;
			this.bitDepth = bitDepth;
			this.channels = channels;
			projectLabel = projLab;
			groupLabel = groupLab;
			hardwareLabel = hardLab;
			description = descLab;
			configure();
			}


		protected void configure()
			{
			long fs;
			try
				{
				fs = raFile.length();
				}
			catch (IOException ex)
				{
				fs = headerSize; // asuume that in case of IOException, the minimum file size is the header size, to avoid problems...
				}
			audioSize = fs - headerSize;
			blockSize = channels * (bitDepth / 8);
			bytesPerSec = rate * blockSize;
			sampleSize = blockSize / channels;
			totalBlocks = audioSize / blockSize;
			}


		public RandomAccessFile getFile()
			{
			return raFile;
			}


		public File getPhisicalFile()
			{
			return phisicalFile;
			}


		@Override
		public void finalize()
			{
			closeFile();
			}


		public String getInfo()
			{
			float seconds;
			int minutes, hours;
			seconds = 8.0f * audioSize / (channels * rate * bitDepth);
			seconds -= (minutes = ((int) seconds) / 60) * 60;
			minutes -= (hours = minutes / 60) * 60;
			String sRate = "" + rate;
			String sBitDepth = "" + bitDepth;
			String sSize = "" + audioSize + headerSize;
			String sAudioSize = "" + audioSize;
			String sTotalBlocks = "" + totalBlocks;
			String sh;
			String sm;
			String ss;
			sh = "" + hours;
			sm = "" + minutes;
			ss = "" + (int) seconds;
			String sTime = sh + "h," + sm + "m," + ss + "s";
			String s = "\tfilename=" + phisicalFile + "\n\t" + "rate=" + sRate + "\n\tchannels=" + (channels == 1 ? "MONO, " : "STEREO") + " bitDepht=" + sBitDepth + "\n\t" + "audioSize=" + sAudioSize + "\n\tfileSize=" + sSize + "\n\ttotalBlocks=" + sTotalBlocks + "\n\ttotalTime=" + sTime + "\n\tprojectLabel=" + projectLabel + "\n\tgroupLabel="
					+ groupLabel + "\n\thardwareLabel=" + hardwareLabel + "\n\tdescriptionLabel=" + description;
			return s;
			}


		public void rewindAll()
			{
			try
				{
				raFile.seek(0);
				}
			catch (IOException ex)
				{
				System.out.println("Error : " + ex);
				}

			}


		public void rewindAudio()
			{
			rewindAll();
			skipHeader();
			}


		public void goBotton()
			{
			try
				{
				long l = raFile.length();
				raFile.seek(l);
				System.out.println("now at : " + raFile.getFilePointer());
				}
			catch (IOException e)
				{
				// TODO Auto-generated catch block
				e.printStackTrace();
				}
			}


		public void skipHeader()
			{
			try
				{
				raFile.seek(headerSize);
				}
			catch (IOException ex)
				{
				System.out.println("Error : " + ex);
				}
			}


		public void closeFile()
			{
			try
				{
				raFile.close();
				}
			catch (IOException ex)
				{
				System.out.println("Error : " + ex);
				}
			}


		public void readHeader() throws IOException
			{
			rewindAll();
			Byter.readMany_U15(raFile, header);
			}


		public void writeHeader() throws IOException
			{
			rewindAll();
			Byter.writeMany_U15(raFile, header);

			}


		public abstract void appendSamplesFromFile(AudioMedia sourceFile) throws IOException;


		public abstract void appendSamplesFromBuffer(byte[] buffer, int bufferSize) throws IOException;


		public abstract AudioMedia copyTo(File newFile) throws IOException;


		public void blockSeek(long blockPos)
			{
			long bp = headerSize + blockPos * blockSize;
			try
				{
				raFile.seek(bp);
				}
			catch (IOException ex)
				{
				System.out.println("Error : " + ex);
				}
			}


		public void byteSeek(long bytePos)
			{
			try
				{
				raFile.seek(bytePos);
				}
			catch (IOException ex)
				{
				System.out.println("Error : " + ex);
				}
			}


		public long blockTell()
			{
			long blockPos = 0;// MIG (ftell(file) - headerSize) / blockSize;
			return blockPos;
			}


		public long byteTell()
			{
			return 0;// MIG ftell(file);
			}


		public int readFragment(byte[] frag) throws IOException
			{
			return raFile.read(frag);
			}


		public int writeFragment(byte[] frag) throws IOException
			{
			return 0; // TODO
			}


		public int readFragment(byte[] frag, int numBytes) throws IOException
			{
			return raFile.read(frag, 0, numBytes);
			}


		public int writeFragment(byte[] frag, int numBytes) throws IOException
			{
			return 0; // TODO
			}


		public abstract void fixAudioSize() throws IOException;


		public final boolean eof()
			{
			// TODO
			return false;
			}


		public final int getBestFragSize()
			{
			return bestFragSize;
			}


		// --- MERGED FROM AUDIO

		/*
		 * SERA QUE  NECESSARIO ?? public void create(String pRootDir, String filename, int pChannels, int pRate, int pSize, String pProjectLabel, String pGroupLabel, String pHardwareLabel, String pDescription) throws IOException { rootDir = pRootDir; pureFilename = filename; PMESG(2, "Creating audio " + pChannels + " channels, " + pRate + " Hz, " +
		 * pSize + " bits"); String absoluteFilename = rootDir + "/" + filename; try { media = new Praf(new File(absoluteFilename)); } catch (IOException ex) { PERROR("Cannot create new file : " + ex); } media.write_header(pChannels, pRate, pSize, 0, pProjectLabel, pGroupLabel, pHardwareLabel, pDescription); try { media.read_header(); } catch
		 * (IOException ex) { System.out.println("Error : " + ex); } }
		 */

		public void buildPeaks() throws PeakException
			{
			if (!peakLess)
				{
				if (peak != null)
					{
					peak = null;
					System.gc();
					}
				try
					{
					peak = new Peak(this);
					}
				catch (PeakException ex)
					{
					PWARN("Cannot create peak for " + get_absolute_filename() + ". Using empty peak strategy.");
					throw ex;
					}
				}
			else
				{
				PWARN("Ignoring peak building... (PeakLess Strategy)");
				}
			}


		public void unbuild()
			{
			closeFile();
			}


		public String get_info()
			{
			String s = "";// MIG file.get_info();
			String s2 = "";// MIG s2.setNum(get_clips_count());
			s = s + "\nClipsCount=" + s2 + "\n";
			return s;
			}


		public int get_clips_count()
			{
			return 0;// MIGparentSong.get_clips_count_for_audio(this);
			}


		public String get_absolute_filename()
			{
			return getPhisicalFile().getPath();
			}


		public String getRootDir()
			{
			return getPhisicalFile().getParent();
			}


		public String getPureName()
			{
			return Util.baseName(phisicalFile.getName());
			}


		protected void dumpHeader()
			{
			PMESG(4, "------ Header for " + phisicalFile + " -------");
			PMESG(4, getInfo());
			PMESG(4, Byter.dump(header));
			}


		public long getFirstAudioBytePos()
			{
			return headerSize;
			}


		public Peak getPeak()
			{
			if (peak != null)
				{
				try
					{
					peak.defineSourceMedia(this);
					}
				catch (PeakException e)
					{
					PERROR("Cannot initialize a Peak object for this audio media : " + e);
					PWARN("Assuming this medias as PeakLess.");
					peakLess = true;
					peak = null;
					}
				}
			else
				{
				PWARN("INTERNAL : A Peak Object was requested for audio media " + this.get_absolute_filename() + " which is set to be peakless.. Maybe it should not be that way...");
				}
			return peak;
			}


		public void setPeak(Peak peak)
			{
			this.peak = peak;
			}


		/** Read samples from media, jumping "jumpStep" bytes after each read. Usefull to read amostrative samples for peak building purposes */
		public void read_samples(byte[] buf, long firstAudioPos, int jumpStep) throws IOException
			{
			int totalSamplesToRead = buf.length;
			int count = 0;
			long scanningBlockPos = firstAudioPos;
			while ((count < totalSamplesToRead) && !eof())
				;
				{
				blockSeek(scanningBlockPos); // FIXME ( ogg files and this next block_seek call) this block_seek is positioning the files always in position 0 no matter what the scanningBlockPos value is It is strange since in ogg seeking test, it worked fine
				byte[] sample = new byte[blockSize];
				raFile.read(sample);
				System.arraycopy(buf, count, sample, 0, 4);
				scanningBlockPos += jumpStep;
				count += blockSize;
				}
			}


		protected void log(String string)
			{
			PMESG(2, "[ AudioMedia " + getPureName() + " ] " + string);

			}


		public void setPeakless(boolean b)
			{
			if ((peakLess) && (peak == null) && !b)
				{
				peak = new Peak();
				peakLess = false;
				}
			else if (b)
				{
				if (peak != null)
					{
					peak = null;
					System.gc();
					peakLess = true;
					}
				}
			}


		public void prepareForReadingSamples(int numSamples)
			{
			if ((numSamples > numSamplesToRead) || (byteBufferForReading == null))
				{
				numSamplesToRead = numSamples;
				if (!clean32BitsSupport)
					{
					byteBufferForReading = ByteBuffer.allocate(numSamplesToRead * sampleSize);
					byteBufferForReading.position(0);
					byteBufferForReading.order(ByteOrder.LITTLE_ENDIAN);
					}
				}
			}


		/**
		 * read and normalizes samples from the media into a float buffer Native 32 Bits PRAF files will read samples much more efficiently than non 32 bits files.
		 */
		public int readSamples(float[] sampleBuffer) throws IOException
			{
			fileChannel.read(byteBufferForReading);
			if (clean32BitsSupport) // yeah .. this audio source is 32 bits , so, just read the whole thing at once , which is much faster
				{
				FloatBuffer floatBufferView = byteBufferForReading.asFloatBuffer();
				floatBufferView.get(sampleBuffer, 0, numSamplesToRead);
				}
			else
				{
				byteBufferForReading.position(0);
				int t = 0;
				switch (bitDepth)
					{
					case Const.DEPTH_8_BITS:
						{
						t = sampleBuffer.length;
						for (int i = 0; i < t; i++)
							sampleBuffer[i] = byteBufferForReading.get(); // TODO NORMALIZE TO -1.0 .. +1.0 float
						break;
						}
					case Const.DEPTH_16_BITS:
						{
						t = sampleBuffer.length / 2;
						for (int i = 0; i < t; i++)
							sampleBuffer[i] = byteBufferForReading.getShort(); // TODO NORMALIZE TO -1.0 .. +1.0 float
						break;
						}
					case Const.DEPTH_24_BITS:
						{
						t = sampleBuffer.length / 3;
						byte b1, b2, b3;
						for (int i = 0; i < t; i++)
							{
							b1 = byteBufferForReading.get();
							b2 = byteBufferForReading.get();
							b3 = byteBufferForReading.get();
							sampleBuffer[i] = b1 * 65536 + b2 * 256 + b3; // REVIEW AND  // TODO NORMALIZE TO -1.0 .. +1.0 float
							}
						break;
						}
					}
				}
			return numSamplesToRead;
			}
	}
